#! /bin/env python3
# SPDX-FileCopyrightText: SIMPLE-Crypto contributors
# SPDX-License-Identifier: Apache-2.0
#
# Copyright SIMPLE-Crypto contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import sys

if sys.platform == "linux":
    FTD2XX = False
else:
    FTD2XX = True  # Set False if using pylibftdi
FTDI_TIMEOUT = 1000  # Timeout for D2XX read/write (msec)

if FTD2XX:
    import sys, time, ftd2xx as ftd
else:
    import sys, time, pylibftdi as ftdi

# Set mode (bitbang / MPSSE)
def set_bitmode(d, bits, mode):
    return (
        d.setBitMode(bits, mode) if FTD2XX else d.ftdi_fn.ftdi_set_bitmode(bits, mode)
    )


# Open device for read/write
def ft_open(idstr=None):
    if FTD2XX:
        if idstr is not None:
            d = ftd.openEx(idstr)
        else:
            d = ftd.open(dev=0)
        d.setTimeouts(FTDI_TIMEOUT, FTDI_TIMEOUT)
    else:
        if idstr is not None:
            d = ftdi.Device(device_id=idstr)
        else:
            d = ftdi.Device(device_index=0)
    return d


# Set SPI clock rate
def set_spi_clock(d, hz):
    div = int((12000000 / (hz * 2)) - 1)  # Set SPI clock
    spi_write(d, (0x86, div % 256, div // 256))


# Read byte data into list of integers
def spi_read(d, nbytes):
    s = d.read(nbytes)
    return [ord(c) for c in s] if type(s) is str else list(s)


# Write list of integers as byte data
def spi_write(d, data):
    s = str(bytearray(data)) if sys.version_info < (3,) else bytes(data)
    return d.write(s)


# Send a read commandto the SPI interface (0xAA)
def spi_read_cmd_bytes(d, nbytes):
    # The input field "nbytes" corresponds to the number of bytes to receive.
    # The preamble and the postamble are built to provide SPI mode 0.
    preamble = [0x80, 0x00, 0x0B, 0x35]
    postamble = [0x80, 0x08, 0x0B]

    read_cmd = list([0xAA])
    # The padding is needed for the SPI since it writes a read command and a zeros
    # to let the slave to send back the requested number of bytes.
    pad = list(np.zeros(nbytes, np.uint8))

    data_package = preamble + [nbytes % 256, nbytes // 256] + read_cmd + pad + postamble
    spi_write(d, data_package)
    temp = spi_read(d, nbytes + 1)
    return temp[1 : nbytes + 1]


# Send a write command to the SPI interface
def spi_write_cmd_bytes(d, data):
    # The preamble and postamble are built to provide SPI mode 0.
    preamble = [0x80, 0x00, 0x0B, 0x11]
    postamble = [0x80, 0x08, 0x0B]

    n = len(data)

    # See document SPI.docx for write commands to the SPI interface.
    write_cmd = list([0xCC])
    spi_write(d, preamble + [n % 256, n // 256] + write_cmd + list(data) + postamble)


# Send the reset command to the Slave (0x99)
def spi_resetSlave(d):
    # The spi_resetSlave command clean the rx_data buffer on the SPI interface.
    # This command does not reset the device, but only the receive buffer of the
    # interface.
    # The preamble and postamble are built to respect the SPI mode 0.
    preamble = [0x80, 0x00, 0x0B, 0x11]
    postamble = [0x80, 0x08, 0x0B]

    # See document SPI.docx for reset commands to the SPI interface.
    reset_cmd = [0x99]

    spi_write(d, preamble + [0, 0] + reset_cmd + postamble)


# Set pins on the FT2232H
def set_ftdi_pinSPI(d):
    # This function set correctly the pins on the FT2232H chip.
    spi_write(d, (0x80, 0x08, 0x0B))
    spi_write(d, (0x82, 0x00, 0x00))

