#! /bin/env python3
# SPDX-FileCopyrightText: SIMPLE-Crypto contributors
# SPDX-License-Identifier: Apache-2.0
#
# Copyright SIMPLE-Crypto contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time
import os

import numpy as np
import functools as ft

import SPIlib as spi
from aes_prng import AESprng

FTDI_ID_DEFAULT=os.environ.get("SAKURA_FTDI_ID")

# New API
from dataclasses import dataclass
@dataclass
class InputByteConfig:
    # Dtype
    TYPE_RANDOM = 0
    TYPE_CONSTANT = 1
    TYPE_SHARE = 2
    # Values
    dtype: int
    value: int
    share_id: int
    @classmethod
    def random(cls) -> 'InputByteConfig':
        return InputByteConfig(InputByteConfig.TYPE_RANDOM, 0, None)
    @classmethod
    def constant(cls, value: int) -> 'InputByteConfig':
        return InputByteConfig(InputByteConfig.TYPE_CONSTANT, value, None)
    @classmethod
    def share(cls, value: int, share_id: int) -> 'InputByteConfig':
        return InputByteConfig(InputByteConfig.TYPE_SHARE, value, share_id)

@dataclass
class InputConfig:
    bytes: list[InputByteConfig]
    
    def __post_init__(self):
        self.__get_shares_infos()

    def __len__(self):
        return len(self.bytes)
    def __getitem__(self, key):
        return self.bytes[key]
    def __add__(self, o: 'InputConfig') -> 'InputConfig':
        return InputConfig(self.bytes + o.bytes)
    def __get_shares_infos(self):
        # Identify the bytes that are encoded as a share
        share_addr = np.where([e.dtype == InputByteConfig.TYPE_SHARE for e in self.bytes])[0]
        shares_infos = {}
        for addr in share_addr:
            bcfg = self.bytes[addr]
            if bcfg.share_id in shares_infos:
                shares_infos[bcfg.share_id].append(addr)
            else:
                shares_infos[bcfg.share_id] = [addr]
        # Validate the fact that at least two shares are used for each indexes
        for shid, addr in shares_infos.items():
            if len(addr)==1:
                raise ValueError("At least two shares should be specified for index {}".format(shid))
        self.shares_infos = shares_infos

# InputConfig([InputByteConfig.random(), InputByteConfig.constant(1), InputByteConfig.share(0, 0), InputByteConfig.share(1, 0), InputByteConfig.share(2, 1), InputByteConfig(3, 1), InputByteConfig.share(4, 0)])
# [a, b, c, d, e, f, g]
# a = uniform
# b = 1
# c, d, g, uniform such that c ^ d ^ g =  0 ^ 1 ^ 4
# e, f, uniform such that e ^ f =  2 ^ 3
@dataclass
class BatchConfig:
    ram: list[int]
    ram_size: int
    flags: list[int]
    nstates: int
    refresh_ram: list[int]
    n_refresh: int
    base: int

    def __post_init__(self):
        # Some validation
        if self.n_refresh != (len(self.refresh_ram)//2):
            raise ValueError("[FATAL ERROR]: each refresh should be configured with two values")

    # Generate the controller configuration based on the abstract configurations
    @classmethod
    def generate(cls, input_configs: list[InputConfig]) -> 'BatchConfig':
        # First of all, verify that all inputs_configs have the same len
        len_inputs_cfg = [len(e) for e in input_configs]
        if len(np.unique(len_inputs_cfg))!=1:
            raise ValueError("The configuration should be of the same length")
        # Second, recover the amounf of rams
        nstates = len(input_configs)
        ram_size = len(input_configs[0])
        # Then, generate the initial rams values by concat. all and generate
        # a constant 0 
        glb_config = ft.reduce(lambda a, b: a+b, input_configs)
        rams = [e.value for e in glb_config]
        flags = ram_size*[0]
        
        # Then generate the randomness generation
        random_byte_status = [e.dtype == InputByteConfig.TYPE_RANDOM for e in glb_config]
        garbage_addr = len(rams) # Address of the garbage byte used for randomness
        addr_rnd = np.where(random_byte_status)[0].tolist()
        if len(addr_rnd)>0:
            rfrsh_pairs_rnd = ft.reduce(lambda a,b: a+b, [[garbage_addr, addr] for addr in addr_rnd])
        else:
            rfrsh_pairs_rnd = []
        
        # Finally, generate the refresh for shares
        rfrsh_pair_rfrsh = []
        for cfgi, cfg in enumerate(input_configs):
            # Recover the shares infos
            shares_info = cfg.shares_infos
            # Iterate over the share_id to generate the refresh
            for _, addrs in shares_info.items():
                # Generate the refresh pair internal to the configuration
                if len(addrs)>2:
                    pairs = ft.reduce(lambda a,b: a+b, [[addrs[i],addrs[(i+1)%len(addrs)]] for i in range(len(addrs))])
                else:
                    pairs = addrs
                # Append with the configuration offset 
                rfrsh_pair_rfrsh += [e+cfgi*ram_size for e in pairs]

        # Concat randomness generation and refresh to get the refresh
        # configuration
        rfrsh_ram = rfrsh_pairs_rnd + rfrsh_pair_rfrsh

        # Return 
        return BatchConfig(
                    rams,
                    ram_size,
                    flags,
                    nstates,
                    tuple(rfrsh_ram),
                    len(rfrsh_ram)//2,
                    ram_size
                )


class Controller:
    # Command opcodes
    WRITE_CMD = 0x22
    CSR_CMD = 0x33
    RUN_CMD = 0x77
    READ_CMD = 0x55
    POLL_CMD = 0xBB
    RESET_CMD = 0xAA
    REFRESH_WRITE0_CMD = 0xCC
    REFRESH_WRITE1_CMD = 0xDD
    READ_CRC_CMD = 0xEE

    # POLL results
    STATUS_WAITRX = 0x5

    # RAM Size (I) (2 bytes)
    I_OFFSET = 0x0000

    # PRNG0 seed
    SEED0_OFFSET = 0x0002
    SEED0_SIZE = 4

    # number of runs
    N_OFFSET = 0x0006
    N_SIZE = 4

    # Flags (F)
    F_OFFSET = 0x0018
    INDIV_F_SIZE = 2 #bits

    # base pointer
    B_OFFSET = 0x000b
    B_SIZE = 2

    # refresh number
    K_OFFSET = 0x000d
    K_SIZE = 2

    # MISC
    MISC_OFFSET = 0x000A
    MISC_SIZE = 0x0001

    # PRNG1 seed
    SEED1_OFFSET = 0x0010
    SEED1_SIZE = 4

    # BYTES CRC
    CRC_BYTES_OFFSET = 0x0014
    CRC_BYTES_SIZE = 2

    # NSTATES BYTES
    NSTATES_OFFSET = 0x0016
    NSTATES_SIZE = 1

    def __init__(
            self, 
            nb_runs, 
            seed0, 
            seed1, 
            ram_size=0, 
            ram=[], 
            flags=[], 
            base=0, 
            n_refresh=0, 
            spi_clock=None, 
            refresh_ram=tuple(), 
            always_reset=False, 
            cmp_crc=False, 
            bytes_crc=None, 
            use_ext_clk_target=False,
            nstates = 1,
            idstr=FTDI_ID_DEFAULT):
        # Only init spi if spi clock specified
        self.set_config(ram_size, ram, nb_runs, flags, seed0, seed1, base, n_refresh, refresh_ram, always_reset, cmp_crc, bytes_crc, use_ext_clk_target, nstates)
        if spi_clock != None:
            self.init_spi(spi_clock, idstr)
            self.reset()
        else:
            print("[WARNING] Controller used in 'offline' mode. No SPI comm will be done.")
            self.dev = None

    def init_spi(self, spi_clock, idstr):
        self.dev = spi.ft_open(idstr)
        spi.set_bitmode(self.dev,0x00,0)
        spi.set_bitmode(self.dev,0x03,2)
        spi.set_spi_clock(self.dev,spi_clock)
        spi.set_ftdi_pinSPI(self.dev)

    def update(self, ram=None, nb_runs=None, flags=None, seed0=None, seed1=None, base=None):
        if ram is not None:
            self.ram = ram
            if self.dev != None:
                self._set_ram()
        if nb_runs is not None:
            self.nb_runs = nb_runs
            if self.dev != None:
                self._set_nb_runs()
        if flags is not None:
            assert len(flags) == self.ram_size
            self.flags = flags
            if self.dev != None:
                self._set_flags()
        if seed0 is not None:
            assert len(seed0) == self.SEED0_SIZE
            self.seed0 = seed0
            if self.dev != None:
                self._set_seed0()
        if seed1 is not None:
            assert len(seed1) == self.SEED1_SIZE
            self.seed1 = seed1
            if self.dev != None:
                self._set_seed1()
        if base is not None:
            self.base = base
            if self.dev != None:
                self._set_base()

    def set_config(self, ram_size, ram, nb_runs, flags, seed0, seed1, base, n_refresh, refresh_ram, always_reset, cmp_crc, bytes_crc, use_ext_clk_target, nstates):
        if flags != None:
            assert len(flags) == ram_size
        assert len(seed0) == self.SEED0_SIZE
        assert len(seed1) == self.SEED1_SIZE
        assert len(refresh_ram) == 2*n_refresh
        if cmp_crc:
            assert bytes_crc!=None
        self.ram_size = ram_size
        self.ram = ram
        self.nb_runs = nb_runs
        self.flags = flags
        self.seed0 = seed0
        self.seed1 = seed1
        self.base = base
        self.n_refresh = n_refresh
        self.refresh_ram = refresh_ram
        self.always_reset = always_reset
        self.cmp_crc = cmp_crc
        if cmp_crc:
            self.bytes_crc = bytes_crc
        else:
            self.bytes_crc = self.ram_size
        self.use_ext_clk_target = use_ext_clk_target
        self.nstates = nstates

    def apply_config(self, batchcfg: BatchConfig):
        self.ram = batchcfg.ram
        self.ram_size = batchcfg.ram_size
        self.flags = batchcfg.flags
        self.nstates = batchcfg.nstates
        self.refresh_ram = batchcfg.refresh_ram
        self.n_refresh = batchcfg.n_refresh
        self.base = batchcfg.base
        if self.dev != None:
            self.send_config()

    def send_config(self):
        self._set_ram_size()
        self._set_ram()
        self._set_nb_runs()
        self._set_seed0()
        self._set_seed1()
        self._set_flags()
        self._set_misc()
        self._set_base()
        self._set_n_refresh()
        self._set_refresh_ram()
        self._set_crc_bytes()
        self._set_nstates_bytes()

    def _write_cmd(self, raw_data):
        """Issue a write command at the low-level SPI protocol."""
        spi.spi_write_cmd_bytes(self.dev, raw_data)

    def _read_cmd(self, n_bytes):
        """Issue a write command at the low-level SPI protocol."""
        return spi.spi_read_cmd_bytes(self.dev, n_bytes)

    def write(self, address, data, cmd=WRITE_CMD):
        """Write the byte data at the given address."""
        a0, a1 = address.to_bytes(2, byteorder="big")
        self._write_cmd([cmd, a0, a1, data])

    def write_block(self, offset, data, cmd=WRITE_CMD):
        """Writes bytes in data in a contiguous segment starting at offset."""
        for i, d in enumerate(data):
            self.write(offset+i, d, cmd)

    def reset(self):
        self._reset()
        self.send_config()

    def _reset(self):
        self._write_cmd([self.RESET_CMD, 0x0, 0x0, 0x0])

    def _set_ram_size(self):
        self.write_block(self.I_OFFSET, self.ram_size.to_bytes(2, byteorder='big'), self.CSR_CMD)

    def _encode_flags(self,f):
        bits_f = len(f)*[0]
        for e_idx, e in enumerate(f):
            bits_f[e_idx] = np.unpackbits(np.array([e],dtype=np.uint8))[8-self.INDIV_F_SIZE:].tolist()
        bf_pack = ft.reduce(lambda a,b: a+b,bits_f)
        return list(np.packbits(bf_pack))

    def _set_flags(self):
        """flags: array of ints (0,1,2) and length ram_size"""
        if (self.flags != None) and (len(self.flags)>0):
            flags_bytes = self._encode_flags(self.flags)#list(np.packbits(self.flags)) # BE
            self.write_block(self.F_OFFSET, flags_bytes, self.CSR_CMD)

    def _set_seed0(self):
        """seed0: list of SEED0_SIZE bytes"""
        self.write_block(self.SEED0_OFFSET, self.seed0, self.CSR_CMD)

    def _set_seed1(self):
        """seed0: list of SEED1_SIZE bytes"""
        self.write_block(self.SEED1_OFFSET, self.seed1, self.CSR_CMD)

    def _set_base(self):
        """base: list of B_SIZE ints"""
        self.write_block(self.B_OFFSET, self.base.to_bytes(self.B_SIZE, byteorder='big'), self.CSR_CMD)

    def _set_n_refresh(self):
        """base: list of K_SIZE ints"""
        self.write_block(self.K_OFFSET, self.n_refresh.to_bytes(self.K_SIZE, byteorder='big'), self.CSR_CMD)

    def _set_misc(self):
        """misc: misc control bits"""
        misc = bytearray(self.MISC_SIZE)
        misc[0] = 0x00
        if self.always_reset:
            misc[0] |= 1
        if self.cmp_crc:
            misc[0] |= 2
        if self.use_ext_clk_target:
            misc[0] |= 4
        self.write_block(self.MISC_OFFSET, misc, self.CSR_CMD)

    def _set_nb_runs(self):
        """nb_runs: int > 0 and < 2**32, or 0 for infinite loop"""
        nb_runs_bytes = self.nb_runs.to_bytes(self.N_SIZE, byteorder="big")
        self.write_block(self.N_OFFSET, nb_runs_bytes, self.CSR_CMD)

    def _set_ram(self):
        """ram array of int of length ram_size"""
        self.write_block(0, self.ram)

    def _set_refresh_ram(self):
        """ram array of int of length ram_size"""
        ref_ram_lsb = [addr & 0xFF for addr in self.refresh_ram]
        ref_ram_msb = [(addr >> 8) & 0xFF for addr in self.refresh_ram]
        self.write_block(0, ref_ram_lsb, cmd=self.REFRESH_WRITE0_CMD)
        self.write_block(0, ref_ram_msb, cmd=self.REFRESH_WRITE1_CMD)

    def _set_crc_bytes(self):
        self.write_block(self.CRC_BYTES_OFFSET, self.bytes_crc.to_bytes(self.CRC_BYTES_SIZE, byteorder='big'), self.CSR_CMD)

    def _set_nstates_bytes(self):
        self.write_block(self.NSTATES_OFFSET, self.nstates.to_bytes(self.NSTATES_SIZE, byteorder='big'), self.CSR_CMD)

    def start_run(self):
        self._write_cmd([self.RUN_CMD, 0x0, 0x0, 0x0])

    def poll(self):
        self._write_cmd([self.POLL_CMD, 0x0, 0x0, 0x0])
        return self._read_cmd(1)[0] & 0x7

    def wait_finished(self, sleep_dur=0.05):
        """Wait until the device finished its operations."""
        while not self.poll() == self.STATUS_WAITRX:
            time.sleep(sleep_dur)

    def read(self, address):
        """Read a byte in ram."""
        self.write(address, 0, cmd=self.READ_CMD)
        return self._read_cmd(1)[0]

    def read_ram(self):
        return [self.read(a) for a in range(self.ram_size)]

    def _crc_raw_to_decimal(self,crc_raw):
        dec = 0
        for i,e in enumerate(crc_raw):
            dec += e << (8*i)
        return dec

    def read_crc(self):
        """Read the CRC"""
        self.write(0, 0, cmd=self.READ_CRC_CMD)
        b0 = self._read_cmd(1)[0]
        self.write(1, 0, cmd=self.READ_CRC_CMD)
        b1 = self._read_cmd(1)[0]
        self.write(2, 0, cmd=self.READ_CRC_CMD)
        b2 = self._read_cmd(1)[0]
        self.write(3, 0, cmd=self.READ_CRC_CMD)
        b3 = self._read_cmd(1)[0]
        return self._crc_raw_to_decimal([b0,b1,b2,b3])

    def predict_ram_inputs(self, prng0=AESprng(), prng1=AESprng()):
        # Amount of bits used for the index 
        LOCALPARAM_BITS_MODULO_MASK = 0b111
        # Amount of bit used from the32 for the randomness
        LOCALPARAM_RND_INDEX_MASK = 0xffffff
        prng0.reseed(self.seed0)
        prng1.reseed(self.seed1)
        ram_state = self.ram
        def _predict_ram_input_run(prng0,prng1):
            #print("predict with",self.base)
            for a0, a1 in zip(self.refresh_ram[::2], self.refresh_ram[1::2]):
                rnd = prng0.nextbyte()
                #print(hex(rnd))
                if a0<len(ram_state):
                    ram_state[a0] ^= rnd
                if a1<len(ram_state):
                    ram_state[a1] ^= rnd
            # Compute a random 32 bits integer to compute random nstates
            first_rnd = prng0.nextbyte()
            second_rnd = prng0.nextbyte()
            third_rnd = prng0.nextbyte()
            fourth_rnd = prng0.nextbyte()
            rnd_nstate = first_rnd << 24
            rnd_nstate += second_rnd << 16
            rnd_nstate += third_rnd << 8
            rnd_nstate += fourth_rnd
            rnd_nstate = rnd_nstate & LOCALPARAM_RND_INDEX_MASK
            state_index = (rnd_nstate % (self.nstates & LOCALPARAM_BITS_MODULO_MASK)) & LOCALPARAM_BITS_MODULO_MASK

            base = state_index * self.base

            #res = (use_base, [prng.nextbyte() if f else self.ram[a+base] for a, f in enumerate(self.flags)])
            res = (state_index, [self.ram[a+base] if f==0 else prng0.nextbyte() if f==1 else prng1.nextbyte() for a,f in enumerate(self.flags)])
            prng0.clear_buffer()
            prng1.clear_buffer()
            return res
        return [_predict_ram_input_run(prng0,prng1) for _ in range(self.nb_runs)][:]

    # Return the string used to depict a word 
    # in the testvector file
    def bytearray2str(self,barray):
        bytestr = ["{:02x}".format(e) for e in barray]
        bytestr.reverse()
        stre = ""
        for bstr in bytestr:
            stre += bstr
        return stre

    # Generate the following batch values sent to the target FPGA
    # and write them in a testvector file for 
    # development/debug purpose
    def generate_testvectors_file(self, filename):
        # Save seeds. The simulation is done
        # such that the internal tate of the controller is not modified
        # after the simulation
        seed0_bk = self.seed0.copy()
        seed1_bk = self.seed1.copy()
        ram_bk = self.ram.copy()
        # Get the inputs
        predicted_batch_ram = self.predict_ram_inputs()
        # Open file and write every cases 
        with open(filename, "w") as f:
            f.write("NEXEC = {}\n".format(len(predicted_batch_ram)))
            for idx, case in enumerate(predicted_batch_ram):
                f.write("CASE: {}\n".format(idx))
                # Print if a reset is required 
                f.write("RST = {}\n".format(int(self.always_reset)))
                # Generate the string to write
                f.write("BUFFER_IN = {}\n".format(self.bytearray2str(case[1])))
        # Re-save the seed
        self.seed0 = seed0_bk
        self.seed1 = seed1_bk
        self.ram = ram_bk

    def __del__(self):
        if self.dev != None:
            self.dev.close()


if __name__ == "__main__":
    # Generate the batch configuration with the new API
    # Here is provided a simple TVLA fixed versus random config

    # Fixed config
    cfg0 = InputConfig([
        InputByteConfig.random(), # RND 
        InputByteConfig.constant(0), # 0
        InputByteConfig.share(np.random.randint(0,256),0), # X0, randomly init, share domain 0
        InputByteConfig.share(np.random.randint(0,256),1), # Y0, randomly init, share domain 1
        InputByteConfig.constant(0), # 0
        InputByteConfig.share(np.random.randint(0,256),0), # X1, randomly init, share domain 0
        InputByteConfig.share(np.random.randint(0,256),1), # Y1, randomly init, share domain 1
        ])
    # Random config
    cfg1 = InputConfig([
        InputByteConfig.random(), # RND 
        InputByteConfig.constant(0), # 0
        InputByteConfig.random(), # X0, random
        InputByteConfig.random(), # Y0, random
        InputByteConfig.constant(0), # 0
        InputByteConfig.random(), # X1, random
        InputByteConfig.random(), # Y1, random
        ])

    bcfg = BatchConfig.generate([cfg0,cfg1])

    # Amount of executions per batch
    nb_runs = 10

    # Seed for the two independant PRNG embedded inside the controller.
    seed0 = np.random.randint(0,256,size=4).tolist()
    seed1 = np.random.randint(0,256,size=4).tolist()

    # Instance of the SAKURA controller 
    c = Controller(
            spi_clock=None,
            nb_runs=nb_runs,
            seed0 = seed0,
            seed1 = seed1,
            always_reset=True,
            )

    # Apply config
    c.apply_config(bcfg)

    # Predict output
    a = c.predict_ram_inputs(AESprng(),AESprng())

    # Generate the testvector file
    testvector_file = "../dev-simu/devtv.tv"
    c.generate_testvectors_file(testvector_file)
