# SPDX-FileCopyrightText: SIMPLE-Crypto contributors
# SPDX-License-Identifier: Apache-2.0
#
# Copyright SIMPLE-Crypto contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import sys
import time
import contextlib

import numpy as np

import picosdk
from picosdk.functions import mV2adc, assert_pico_ok
from picosdk.PicoDeviceEnums import picoEnum as ps6000aEnums
from picosdk.ps5000a import ps5000a
from picosdk.ps6000a import ps6000a

from math import ceil

class ChannelConfig:
    def __init__(self,name,
            coupling,
            vrange,
            enabled=True,
            voffset=0):

        """vrange and voffset in [V]"""
        self.name = name
        self.enabled = enabled
        self.coupling = coupling
        self.vrange = vrange
        self.voffset = voffset

class PicoTriggerInfo(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
            ("status", ctypes.c_int32),
            ("segmentIndex", ctypes.c_uint64),
            ("triggerIndex", ctypes.c_uint64),
            ("triggerTime", ctypes.c_double),
            ("timeUnits", ctypes.c_int32),
            ("missedTriggers", ctypes.c_uint64),
            ("timeStampCounter", ctypes.c_uint64)
            ]

class SCAScope():
    ## TODO: check that this value is sane
    # Rationale: if cutoff freq is 1 Hz (to be checked), then time constant is
    # 1/2pi s -> if we wait 2s, attenutation of DC is > 100 dB
    DC_DECAP_LOAD_TIME = 2 # s
    WAIT_MSG = "Waiting {:.1g}s for scope DC decoupling capacitor loading..."
    # Amount of bytes use to encode a trigger info structure (low-level)
    API_TRIGGER_INFO_BYTES = 48

    def __ResolutionString(self, model, resolution_bits):
        if model=="PS5000A":
            return "PS5000A_DR_{}BIT".format(resolution_bits)
        elif model=="PS6000A":
            return "PICO_DR_{}BIT".format(resolution_bits)
        else:
            raise ValueError("model not handled")

    def __AcqConfiguration(self,model,
            target_resolution,target_sfreq, target_samples):
        if model=="PS5000A":
            # In this case the down sampling is automatically handled by the 
            # firmware. (TODO check to be sure)
            cfg_DOWNSAMPLE_RATIO_MODE = "PS5000A_RATIO_MODE_NONE"
            cfg_DOWNSAMPLE_RATIO = 1
            cfg_sfreq = target_sfreq
            cfg_resolution = self.__ResolutionString(model, target_resolution)
            cfg_nsamples = target_samples
        elif model=="PS6000A":
            if target_resolution > 12:
                delta_res_max = target_resolution - 12
                cfg_DOWNSAMPLE_RATIO_MODE = "PICO_RATIO_MODE_AVERAGE"
                cfg_DOWNSAMPLE_RATIO = 2**delta_res_max
                cfg_sfreq = target_sfreq*cfg_DOWNSAMPLE_RATIO
                cfg_resolution = self.__ResolutionString(model, 12)
                cfg_nsamples = target_samples*cfg_DOWNSAMPLE_RATIO
                print("Oversampling ({}Hz) with downsampling ({} ratio) configuration required".format(cfg_sfreq,cfg_DOWNSAMPLE_RATIO))
            else:
                cfg_DOWNSAMPLE_RATIO_MODE = "PICO_RATIO_MODE_RAW"
                cfg_DOWNSAMPLE_RATIO = 1
                cfg_sfreq = target_sfreq
                cfg_resolution = self.__ResolutionString(model, target_resolution)
                cfg_nsamples = target_samples
        else: 
            raise ValueError("Model not handled")
        return (
                cfg_DOWNSAMPLE_RATIO_MODE,
                cfg_DOWNSAMPLE_RATIO,
                cfg_sfreq,
                cfg_resolution,
                cfg_nsamples
                )
            
    def __init__(self,
            model = "PS5000A",
            #resolution="PS5000A_DR_12BIT", # bits
            resolution=12, # bits
            n_samples=1000,
            sample_freq=500e6, # Hz
            batch_size=1,
            channels = [ChannelConfig(
                name="PS5000A_CHANNEL_A",
                coupling="PS5000A_AC",
                vrange="PS5000A_10MV",
                enabled=True)],
            trig_delay=0,
            timeout_ms=0,
            serialNumber = None,
            dummy_ac = True,
            clkgen_freq = None, # Hz
            clkgen_amp = 2, # V
            clkgen_off = 0, # V
            trigSrc = 'AUX',
            trigThreshmV = 1700,
            ):

        self.overflow = None
        self.data = None

        self.model = model
        self.channels = channels

        # Fetch some configuration
        (cfg_DOWNSAMPLE_RATIO_MODE, cfg_DOWNSAMPLE_RATIO,
                cfg_sfreq, cfg_resolution, cfg_nsamples) = self.__AcqConfiguration(
                        model,
                        resolution,
                        sample_freq,
                        n_samples
                        )

        self.resolution_bits = resolution

        if self.model == "PS5000A":
            self.trigSrc = ps5000a.PS5000A_CHANNEL["PS5000A_EXTERNAL"]
            self.direction = ps5000a.PS5000A_THRESHOLD_DIRECTION["PS5000A_RISING"]
            self.DOWNSAMPLE_RATIO_MODE = ps5000a.PS5000A_RATIO_MODE[cfg_DOWNSAMPLE_RATIO_MODE]
            self.DOWNSAMPLE_RATIO = cfg_DOWNSAMPLE_RATIO
            self.resolution = ps5000a.PS5000A_DEVICE_RESOLUTION[cfg_resolution]
        elif self.model == "PS6000A":
            # Configure the trigger source. 
            if trigSrc == "AUX":
                self.trigSrc = "PICO_TRIGGER_AUX"
                self.trigRange = 2500
            elif trigSrc in ['A','B','C','D']:
                self.trigSrc = "PICO_CHANNEL_{}".format(trigSrc)
                self.trigRange = None
                for c in channels:
                    if c.name == self.trigSrc:
                        self.trigRange = self.decode_PS5000_RANGE_mV(c.vrange)
                if self.trigRange == None:
                    raise ValueError("Trigger vrange not found.")
            else:
                raise ValueError("TrigSrc should be either 'A', 'B', 'C', 'D' or 'AUX'.")
            self.direction = "PICO_RISING"
            self.DOWNSAMPLE_RATIO_MODE = ps6000aEnums.PICO_RATIO_MODE[cfg_DOWNSAMPLE_RATIO_MODE]
            self.DOWNSAMPLE_RATIO = cfg_DOWNSAMPLE_RATIO
            self.BANDWIDTH = ps6000aEnums.PICO_BANDWIDTH_LIMITER["PICO_BW_FULL"]
            self.resolution = ps6000aEnums.PICO_DEVICE_RESOLUTION[cfg_resolution]
            self.DATA_TYPE = ps6000aEnums.PICO_DATA_TYPE["PICO_INT16_T"]

        print('Starting scope... ', end='', file=sys.stderr, flush=True)
        self.chandle = ctypes.c_int16()

        timeIndisposedMs = ctypes.c_double()
        self.timeIndisposedMs = ctypes.byref(timeIndisposedMs)
        if self.model == "PS5000A":
            m = ps5000a.ps5000aOpenUnit(ctypes.byref(self.chandle),
                serialNumber, self.resolution)

            nchannels = 2
            try:
                assert_pico_ok(m)
            except: # PicoNotOkError:
                powerStatus = m
                if powerStatus == 286:
                    m = ps5000a.ps5000aChangePowerSource(self.chandle, m)
                elif powerStatus == 282:
                    m = ps5000a.ps5000aChangePowerSource(self.chandle, m)
                else:
                    raise
                assert_pico_ok(m)


        elif self.model == "PS6000A":
            m = ps6000a.ps6000aOpenUnit(ctypes.byref(self.chandle),
                serialNumber, self.resolution)
            nchannels = 4
            assert_pico_ok(m)

        print('done.', file=sys.stderr)

        # closing all the channels:
        for i in range(nchannels):
            self.setChannel(channel=i,
                    coupling =0,VRange=0,VOffset=0,enabled=False)

        # apply requested settings for channels
        for channel_config in channels:
            self.setChannel(
                    channel=channel_config.name,
                    coupling=channel_config.coupling,
                    VRange=channel_config.vrange,
                    VOffset=channel_config.voffset,
                    enabled=channel_config.enabled,
                    )

        (sampling_freq, max_samples) = self.setSamplingFrequency(
                sampleFreq=cfg_sfreq,
                noSamples=cfg_nsamples,
                oversample=0,
                segmentIndex=0
                )

        # Verify if the frequency is reachable
        #if model=="PS6000A":
        #    if sampling_freq>625e6:
        #        raise ValueError("SAmpling Frequency unreachable")

        self.noSamples = cfg_nsamples
        self.sampling_freq = sampling_freq
        self.max_samples = max_samples

        # Start waveform generator
        self.setSigGen(clkgen_amp,clkgen_off,clkgen_freq)


        # apply channel AC/DC settings in HW
        if dummy_ac:
            self._dummy_acquisition()
        self._initialized = False
        self._init_start_time = time.time()


        self._setSimpleTrigger(
                trigSrc=self.trigSrc,
                timeout_ms=timeout_ms, # us (and not ms !), wait time before forced trigger. 0 to disable.
                enabled=True,
                threshold_mV = trigThreshmV,
                direction=self.direction,delay=trig_delay*self.DOWNSAMPLE_RATIO
                )

        # TODO fix this function in picoscope library
        #if batch_size > self.getMaxMemorySegments():
        #    raise ValueError("Too many batches")

        max_samples = self.memorySegments(batch_size)
        if max_samples < cfg_nsamples:
            raise ValueError("Too many batches for the number of samples required.")

        self._setNoOfCaptures(batch_size)

        print("Sampling Freq: {:g} MHz".format(self.sampling_freq/1E6))
        print("maxSamples: {}".format(self.max_samples))
        print("noSamples: {}".format(self.noSamples))
        print("resolution: {}".format(self.resolution))


    def decode_PS5000_RANGE_mV(self, str_range):
        valuesp = str_range.split('_')[1][:-1]
        # Check if last char of the str is a char 'M', meaning that the digit is mv 
        if valuesp[-1] == "M":
            digitv = int(valuesp[:-1])
            return digitv
        else:
            digitv = int(valuesp)
            return 1000 * digitv

    def setSigGen(self, amplitude_V, offset_V, freq_hz):
        if freq_hz is not None and freq_hz != 0.0:
            amp_p2p_V = amplitude_V 
            amp_p2p_uV = amp_p2p_V * 1e6
            off_uV = offset_V * 1e6
            offset_uV = amp_p2p_uV / 2.0 + off_uV
            offset_V = offset_uV / 1e6
            if freq_hz < 100:
                raise ValueError("Frequence should be bigger than 100Hz")
            effective_freq = int(freq_hz) 
            print("Start square clock generation of {}Hz [offset:{} ; P2P:{}]".format(effective_freq,offset_V,amp_p2p_V))

            if self.model=="PS5000A":
                m = ps5000a.ps5000aSetSigGenBuiltInV2(
                        self.chandle,
                        offset_uV,
                        amp_p2p_uV,
                        ctypes.c_int32(1), # SQUARE
                        effective_freq,
                        effective_freq,
                        0,
                        0,
                        ctypes.c_int32(0), # UP
                        ctypes.c_int32(0), # ES OFF
                        0,
                        0,
                        ctypes.c_int32(0), # PS5000A_SIGGEN_RISING,
                        ctypes.c_int32(0), # PS5000A_SIGGEN_NONE,
                        0
                        )
                assert_pico_ok(m)
            elif self.model=="PS6000A":
                bufferlength = 1000
                buffer = (ctypes.c_int16 * bufferlength)()
                # Sig waveform
                m = ps6000a.ps6000aSigGenWaveform(
                        self.chandle, 
                        ps6000aEnums.PICO_WAVE_TYPE["PICO_SQUARE"], 
                        ctypes.byref(buffer), 
                        bufferlength
                        )
                assert_pico_ok(m)
                # Sig range
                m = ps6000a.ps6000aSigGenRange(
                        self.chandle, 
                        amp_p2p_V, 
                        offset_V
                        )
                assert_pico_ok(m)
                # Sig duty cycle
                m = ps6000a.ps6000aSigGenWaveformDutyCycle(
                        self.chandle, 
                        50
                        )
                assert_pico_ok(m)
                # Sig freq
                m = ps6000a.ps6000aSigGenFrequency(
                        self.chandle, 
                        effective_freq
                        ) 
                assert_pico_ok(m)
                # Apply settings
                m = ps6000a.ps6000aSigGenApply(
                        self.chandle,
                        1, # Sig gen enable
                        0, # Sweep enabled
                        0, # trig enabled
                        1, # automatic clock optimization enabled
                        0, # override aitomaic clock and prescale
                        ctypes.byref(ctypes.c_int16(effective_freq)),
                        None,
                        None,
                        None
                        )
                assert_pico_ok(m)
            else:
                print("=====> WARNING: clock generation not handled.")
                # TODO: 


    def setChannel(self,channel,coupling,VRange,VOffset,enabled):
        if self.model == "PS5000A":
            if not isinstance(channel,int):
                channel = ps5000a.PS5000A_CHANNEL[channel]
            if not isinstance(coupling,int):
                coupling = ps5000a.PS5000A_COUPLING[coupling]
            if not isinstance(VRange,int):
                VRange = ps5000a.PS5000A_RANGE[VRange]
            m = ps5000a.ps5000aSetChannel(self.chandle,channel,enabled,coupling,VRange,VOffset)
            assert_pico_ok(m)
        elif self.model == "PS6000A":
            if not isinstance(channel,int):
                channel = ps6000aEnums.PICO_CHANNEL[channel]
            if not isinstance(coupling,int):
                coupling = ps6000aEnums.PICO_COUPLING[coupling]
            if not isinstance(VRange,int):
                #TODO move range to PS6000
                VRange = ps5000a.PS5000A_RANGE[VRange]

            if enabled:
                m = ps6000a.ps6000aSetChannelOn(self.chandle,
                        channel,coupling,VRange,VOffset,self.BANDWIDTH)
            else:
                m = ps6000a.ps6000aSetChannelOff(self.chandle,channel)
            assert_pico_ok(m)


    def setSamplingFrequency(self,sampleFreq,
                noSamples,
                oversample,
                segmentIndex):

        if self.model == "PS5000A":
            # testing many timebase and find the closed to the desired freq.
            timeIntervalns = ctypes.c_float()
            returnedMaxSamples = ctypes.c_int32()
            freq = [];tb = []
            timebase = 0 ;tmp = sampleFreq + 1;
            while tmp > sampleFreq:
                try:
                    m = ps5000a.ps5000aGetTimebase2(self.chandle, timebase, noSamples,
                            ctypes.byref(timeIntervalns),
                            ctypes.byref(returnedMaxSamples), segmentIndex)
                    assert_pico_ok(m)
                except picosdk.errors.PicoSDKCtypesError: # oups, I have tried a wrong timebase
                    timebase +=1
                else:
                    tmp = 1/(timeIntervalns.value*1E-9)
                    freq.append(tmp)
                    tb.append(timebase)
                    timebase += 1

            timebase = tb[min(list(range(0,len(tb))),key=lambda x:abs(freq[x]-sampleFreq))]
            m = ps5000a.ps5000aGetTimebase2(self.chandle, timebase, noSamples,
                    ctypes.byref(timeIntervalns),
                    ctypes.byref(returnedMaxSamples), segmentIndex)
            assert_pico_ok(m)

            freq_c = (1/(timeIntervalns.value*1E-9))
            if abs(freq_c - sampleFreq) > .1:
                raise Exception("Provided frequency is not available ",sampleFreq,"possible frequencies",freq)
            self.timebase = timebase
            return freq_c , returnedMaxSamples.value

        elif self.model == "PS6000A":
            # testing many timebase and find the closed to the desired freq.
            timeIntervalns = ctypes.c_double()
            returnedMaxSamples = ctypes.c_int64()
            freq = [];tb = []
            timebase = 0 ;tmp = sampleFreq + 1;
            while tmp > sampleFreq:
                try:
                    m = ps6000a.ps6000aGetTimebase(self.chandle, timebase, noSamples,
                            ctypes.byref(timeIntervalns),
                            ctypes.byref(returnedMaxSamples), segmentIndex)
                    assert_pico_ok(m)
                except picosdk.errors.PicoSDKCtypesError: # oups, I have tried a wrong timebase
                    timebase +=1
                else:
                    tmp = 1/(timeIntervalns.value*1E-9)
                    freq.append(tmp)
                    tb.append(timebase)
                    timebase += 1

            timebase = tb[min(list(range(0,len(tb))),key=lambda x:abs(freq[x]-sampleFreq))]
            m = ps6000a.ps6000aGetTimebase(self.chandle, timebase, noSamples,
                    ctypes.byref(timeIntervalns),
                    ctypes.byref(returnedMaxSamples), segmentIndex)
            assert_pico_ok(m)

            freq_c = (1/(timeIntervalns.value*1E-9))
            if abs(freq_c - sampleFreq) > .1:
                raise Exception("Provided frequency is not available ",sampleFreq,"possible frequencies",freq)
            self.timebase = timebase
            return freq_c , returnedMaxSamples.value

    def _dummy_acquisition(self):
        """A 1 trace, immediately triggered acquisition to force
        (dis)connection of DC decoupling capacitor.
        Wait a few seconds after this !
        """
        self._setNoOfCaptures(1)
        self._setSimpleTrigger(
                trigSrc=self.trigSrc,
                timeout_ms=1, # us (and not ms !), wait time before forced trigger. 0 to disable.
                enabled=True,
                threshold_mV = 1000,
                direction=self.direction,delay=0
                )
        self.runBlock()
        self.waitReady()

    def _setNoOfCaptures(self,n):
        if self.model == "PS5000A":
            m = ps5000a.ps5000aSetNoOfCaptures(self.chandle,n)
        elif self.model == "PS6000A":
            m = ps6000a.ps6000aSetNoOfCaptures(self.chandle,n)

        assert_pico_ok(m)

    def _setSimpleTrigger(self,trigSrc,timeout_ms,enabled,direction,threshold_mV,delay):
        if self.model == "PS5000A":
            if not isinstance(trigSrc,int):
                trigSrc = ps5000a.PS5000A_CHANNEL[trigSrc]
            if not isinstance(direction,int):
                direction = ps5000a.PS5000A_THRESHOLD_DIRECTION[direction]

            maxADC = ctypes.c_int16()
            m = ps5000a.ps5000aMaximumValue(self.chandle, ctypes.byref(maxADC))
            assert_pico_ok(m)

            threshold = threshold_adc = int((threshold_mV / (1000 * 20)) *
                                2**16)
            m = ps5000a.ps5000aSetSimpleTrigger(self.chandle,
                enabled,
                trigSrc,
                threshold,
                direction,
                delay, timeout_ms)
        elif self.model == "PS6000A":
            trigSrc_str = str(trigSrc)
            trigSrc = ps6000aEnums.PICO_CHANNEL[trigSrc]
            direction = ps6000aEnums.PICO_THRESHOLD_DIRECTION[direction]

            maxADC = ctypes.c_int16()
            minADC = ctypes.c_int16()
            m = ps6000a.ps6000aGetAdcLimits(self.chandle,
                    self.resolution,
                    ctypes.byref(minADC),
                    ctypes.byref(maxADC))
            assert_pico_ok(m)

            if trigSrc_str == "PICO_TRIGGER_AUX":
                # In the case of auxiliary port selected, the threshold is fixed to 1.25V with a range of 2.5V. 
                # This is fixed as detailled in the documentation Picoscope6000E Series datashett, page 33 in Auxiliary trigger
                threshold = 0
                pthres_mV = 0
                threshold_ctypes = ctypes.c_int16(0)
                print("Trigger configured on channel {} ati 1250mV (fixed)".format(trigSrc_str,pthres_mV,threshold))
            else:
                # Verify that the trigger value is in the range
                if threshold_mV > self.trigRange:
                    raise ValueError("Unreachable trigger threshold considering the configured range")

                threshold = int(round((threshold_mV * maxADC.value)/self.trigRange))
                pthres_mV = threshold_mV
                threshold_ctypes = ctypes.c_int16(threshold)
                print("Trigger configured on channel {} at {}mV (ADC value {})".format(trigSrc_str,pthres_mV,threshold))

            m = ps6000a.ps6000aSetSimpleTrigger(
                self.chandle,
                enabled,
                trigSrc,
                threshold_ctypes,
                direction,
                delay, timeout_ms)
        assert_pico_ok(m)

    def runBlock(self,pretrig=0.0,segmentIndex=0):
        nSamples = self.noSamples
        nSamples_pretrig = int(round(nSamples*pretrig))
        if self.model == "PS5000A":
            timeIndisposedMs = ctypes.c_int32()
            m = ps5000a.ps5000aRunBlock(self.chandle,
                    nSamples_pretrig,nSamples-nSamples_pretrig,
                    self.timebase,
                    ctypes.byref(timeIndisposedMs),
                    segmentIndex,None,None)
        elif self.model == "PS6000A":
            m = ps6000a.ps6000aRunBlock(self.chandle,
                    nSamples_pretrig,nSamples-nSamples_pretrig,
                    self.timebase,
                    self.timeIndisposedMs,
                    segmentIndex,None,None)
        assert_pico_ok(m)

    def waitReady(self, spin_delay=0.01):
        """Block until the scope is ready."""
        while not self.isReady():
            time.sleep(spin_delay)
    def isReady(self):
        ready = ctypes.c_int16()
        if self.model == "PS5000A":
            m = ps5000a.ps5000aIsReady(self.chandle, ctypes.byref(ready))
        elif self.model == "PS6000A":
            m = ps6000a.ps6000aIsReady(self.chandle, ctypes.byref(ready))

        assert_pico_ok(m)
        return ready

    def ensure_initialized(self):
        if not self._initialized:
            while True:
                remaining_wait = self._init_start_time + self.DC_DECAP_LOAD_TIME - time.time()
                if remaining_wait < 0:
                    break
                print(self.WAIT_MSG.format(remaining_wait), file=sys.stderr)
                time.sleep(remaining_wait) # may sleep for less time than asked
            self.initialized = True

    def getDataRawBulk(self,data,channel):

        """Get data recorded in block mode."""

        # set up each row in the data array as a buffer for one of
        # the memory segments in the scope
        if self.data is None:
            if data.dtype != np.int16:
                raise Exception("Data should have int16 dtype, currently ",data.dtype)

            for i in range(self.noSegments):
                self._lowLevelSetDataBuffer(channel,data[i],i)
            self.data = None
        else:
            if self.data.__array_interface__["data"] != data.__array_interface__["data"]:
                raise Exception("Data is not the one that has been initilialized")

        if self.overflow is None:
            self.overflow = np.ascontiguousarray(np.zeros(self.noSegments, dtype=np.int16))


        self._lowLevelGetValuesBulk(noSamples=self.noSamples,
                    fromSegment=0,
                    toSegment=self.noSegments-1,
                    overflow=self.overflow)

        return (self.data,self.overflow)

    def get_trigger_infos(self):
        ### Frist try, should work but failing ...
        triggerInfo = (PicoTriggerInfo * self.noSegments)()

        # Get trigger infos
        m = ps6000a.ps6000aGetTriggerInfo(
                    self.chandle,
                    ctypes.byref(triggerInfo),
                    0,
                    self.noSegments
                )

        # Check return status
        assert_pico_ok(m)

        # Decode and return
        return triggerInfo
        
    def trigger_offset_as_trace_shift(self): 
        times = np.zeros((self.noSegments,), dtype=np.int64)
        timeUnits = np.zeros((self.noSegments,), dtype=np.int32)
        self.times = times
        self.timeUnits = timeUnits
        
        self._lowLevelGetValuesTriggerTimeOffsetBulk(fromSegment=0,toSegment=self.noSegments-1, times=times, timeUnits=timeUnits)
        tu = ps6000aEnums.PICO_TIME_UNITS
        tu_ref = np.zeros((len(tu),), dtype=np.float64)
        for n, v in [
                ('PICO_FS', 1e-15),
                ('PICO_PS', 1e-12),
                ('PICO_NS', 1e-9),
                ('PICO_US', 1e-6),
                ('PICO_MS', 1e-3),
                ('PICO_S', 1e0),
                ]:
            tu_ref[tu[n]] = v
        offsets = times * tu_ref[timeUnits]
        return offsets*self.sampling_freq

    def useAuxRefClock(self):
        # 2 ==   PS6000_FREQUENCY_10MHZ
        # threshold: not clear what the unit is, either
        # - -1V=-32k, +1V=+32k
        # - or threshold is in mV...
        self._lowLevelSetExternalClock(2, 1000)


    def _lowLevelSetExternalClock(self, frequency, threshold):
        if self.model == "PS6000A":
            m = ps6000a.ps6000aSetExternalClock(
                    self.chandle,
                    frequency,
                    threshold,
                    )
            assert_pico_ok(m)
        else:
            raise NotImplemented()


    def _lowLevelGetValuesTriggerTimeOffsetBulk(self, fromSegment, toSegment, times, timeUnits):
        # times should be int64, timeUnits int32.
        if self.model == "PS6000A":
            assert ctypes.sizeof(ctypes.c_int) == 4 # That should work on all relevant platforms.
            print(times.dtype)
            print(timeUnits.dtype)
            m = ps6000a.ps6000aGetValuesTriggerTimeOffsetBulk(
                    self.chandle,
                    times.ctypes.data_as(ctypes.POINTER(ctypes.c_int64)),
                    timeUnits.ctypes.data_as(ctypes.POINTER(ctypes.c_int)),
                    ctypes.c_uint64(fromSegment),
                    ctypes.c_uint64(toSegment),
                    )
            assert_pico_ok(m)
        else:
            raise NotImplemented()


    def _lowLevelSetDataBuffer(self,channel,data,segment):
        if self.model == "PS5000A":
            dataPtr = data.ctypes.data_as(ctypes.POINTER(ctypes.c_int16))
            if not isinstance(channel,int):
                channel = ps5000a.PS5000A_CHANNEL[channel.name]
            m = ps5000a.ps5000aSetDataBuffer(self.chandle,
                    channel,
                    dataPtr,
                    len(data),
                    segment,
                    self.DOWNSAMPLE_RATIO_MODE)

        elif self.model == "PS6000A":
            dataPtr = data.ctypes.data_as(ctypes.POINTER(ctypes.c_int16))
            if not isinstance(channel,int):
                channel = ps6000aEnums.PICO_CHANNEL[channel.name]

            action = ps6000aEnums.PICO_ACTION["PICO_ADD"]
            m = ps6000a.ps6000aSetDataBuffer(self.chandle,
                    channel,
                    dataPtr,
                    len(data),
                    self.DATA_TYPE,
                    segment,
                    self.DOWNSAMPLE_RATIO_MODE,
                    action)

        assert_pico_ok(m)

    def _checkRetValueAmount(self,noSamples_ret):
        #return noSamples_ret==ceil(self.noSamples/self.DOWNSAMPLE_RATIO) 
        if self.model == "PS5000A":
            return noSamples_ret.value == noSamples
        elif self.model == "PS6000A":
            return noSamples_ret==ceil(self.noSamples/self.DOWNSAMPLE_RATIO)


    def _lowLevelGetValuesBulk(self,noSamples, fromSegment, toSegment, overflow):
        if self.model == "PS5000A":
            noSamples_ret = ctypes.c_int32(noSamples)
            m = ps5000a.ps5000aGetValuesBulk(
                    self.chandle,
                    ctypes.byref(noSamples_ret),
                    fromSegment,
                    toSegment,
                    self.DOWNSAMPLE_RATIO,
                    self.DOWNSAMPLE_RATIO_MODE,
                    overflow.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)))

            if noSamples_ret.value != noSamples:
                raise Exception("Less samples recorded. Required %d and received %d"%(noSamples,noSamples_ret.value))

        elif self.model == "PS6000A":
            noSamples_ret = ctypes.c_int64(noSamples)
            m = ps6000a.ps6000aGetValuesBulk(
                    self.chandle,
                    0,
                    ctypes.byref(noSamples_ret),
                    fromSegment,
                    toSegment,
                    self.DOWNSAMPLE_RATIO,
                    self.DOWNSAMPLE_RATIO_MODE,
                    overflow.ctypes.data_as(ctypes.POINTER(ctypes.c_int16)))
            if not(self._checkRetValueAmount(noSamples_ret.value)):
                raise Exception("Less samples recorded. Required %d and received %d"%(noSamples,noSamples_ret.value))
        assert_pico_ok(m)

    def start_acquire(self):
        self.ensure_initialized()
        self.runBlock(pretrig=0) # no samples before trigger

    def result_acquire(self, dest,channel, allow_overflow=False):
        """Warning: returns the same buffer at every call."""
        self.waitReady() # TODO replace this with non-busy wait and callback
        (retd,retovfl) = self.getDataRawBulk(data=dest,channel=channel)
        # Check the overflow -> raiseValueError in that case
        am_ovfl = np.sum(retovfl)
        if not(allow_overflow) and (retovfl != 0).any():
            raise Exception("Overflow(s) occured during the acquisition!")
        return (retd, retovfl)

    def _lowLevelGetValuesOverlappedBulk(self,noSamples, fromSegment, toSegment, overflow):
        if self.model == "PS5000A":
            noSamples_ret = ctypes.c_int32(noSamples)
            m = ps5000a.ps5000aGetValuesOverlappedBulk(
                    self.chandle,
                    0,
                    ctypes.byref(noSamples_ret),
                    self.DOWNSAMPLE_RATIO,
                    self.DOWNSAMPLE_RATIO_MODE,
                    fromSegment,
                    toSegment,
                    overflow.ctypes.data_as(ctypes.POINTER(ctypes.c_int16))
                    )
        elif self.model == "PS6000A":
            noSamples_ret = ctypes.c_int64(noSamples)
            m = ps6000a.ps6000aGetValuesOverlapped(
                    self.chandle,
                    0,
                    ctypes.byref(noSamples_ret),
                    self.DOWNSAMPLE_RATIO,
                    self.DOWNSAMPLE_RATIO_MODE,
                    fromSegment,
                    toSegment,
                    overflow.ctypes.data_as(ctypes.POINTER(ctypes.c_int16))
                    )
        assert_pico_ok(m)
        return noSamples_ret

    def _lowLevelStopUsingGetValuesOverlapped(self):
        if self.model == "PS5000A":
            pass # no such function for the 5000a API
        elif self.model == "PS6000A":
            m = ps6000a.ps6000aStopUsingGetValuesOverlapped(self.chandle)
            assert_pico_ok(m)

    @contextlib.contextmanager
    def overlapped_acquisitions(self):
        overflow = np.zeros(self.noSegments, dtype=np.int16)
        data = [np.zeros([self.noSegments,self.noSamples//self.DOWNSAMPLE_RATIO],dtype=np.int16) for _ in self.channels]
        for d, channel in zip(data, self.channels):
            for i in range(self.noSegments):
                self._lowLevelSetDataBuffer(channel,d[i,:],i)
        toSegment = ctypes.c_uint64(self.noSegments-1)
        self._noSamples_ret = self._lowLevelGetValuesOverlappedBulk(self.noSamples, 0, self.noSegments-1, overflow)
        self._data_overlapped = data
        self._overflow_overlapped = overflow
        yield self
        self._lowLevelStopUsingGetValuesOverlapped()

    def overlapped_get(self, allow_overflow=False):
        self.waitReady()
        self._checkRetValueAmount(self._noSamples_ret)
        if not allow_overflow and (self._overflow_overlapped != 0).any():
            raise ValueError("Overflow in picoscope acquisition.")
        return self._data_overlapped

    def memorySegments(self,n):
        if self.model == "PS5000A":
            nMaxSamples = ctypes.c_int32()
            m = ps5000a.ps5000aMemorySegments(self.chandle,
                        n,
                        ctypes.byref(nMaxSamples))
        elif self.model == "PS6000A":
            nMaxSamples = ctypes.c_int64()
            m = ps6000a.ps6000aMemorySegments(self.chandle,
                        n,
                        ctypes.byref(nMaxSamples))

        assert_pico_ok(m)

        self.noSegments = n
        return nMaxSamples.value

    def __del__(self):
        if self.model == "PS6000A":
            m = ps6000a.ps6000aCloseUnit(self.chandle)
            assert_pico_ok(m)
        elif self.model == "PS5000A":
            m = ps5000a.ps5000aCloseUnit(self.chandle)
            assert_pico_ok(m)

if __name__ == "__main__":
    model = "PS5000A"
    if model == "PS5000A":
        channelA = ChannelConfig(
                    name="PS5000A_CHANNEL_A",
                    coupling="PS5000A_AC",
                    vrange="PS5000A_10MV",
                    enabled=True)
        resolution = "PS5000A_DR_8BIT"
        sample_freq = 1E9
    else:
        channelA = ChannelConfig(
                    name="PICO_CHANNEL_A",
                    coupling="PICO_AC",
                    vrange="PS5000A_10MV",
                    enabled=True)
        resolution = "PICO_DR_10BIT"
        sample_freq = 5E9

    scope = SCAScope(
            channels=[channelA],
            batch_size=1,
            n_samples=1000,
            timeout_ms=1,
            model=model,
            sample_freq=sample_freq,
            resolution=resolution,
            clkgen_freq=10000,
            clkgen_amp=2
            )

    scope.start_acquire()

    data = np.zeros((scope.noSegments,scope.noSamples),dtype=np.int16)
    scope.result_acquire(data,channelA)
    print(data)
