--------------------------------------------------------------------------------
--
--   FileName:         <name>
--   Dependencies:     none
--   Design Software:  Quartus II 64-bit Version 13.0.1 Build 273 Web Edition
--	  Author: 			  Davide Bellizia davide.bellizia@uclouvain.be
--	  IP:					  SPI Slave Interface
--
--   HDL CODE IS PROVIDED "AS IS."  DIGI-KEY EXPRESSLY DISCLAIMS ANY
--   WARRANTY OF ANY KIND, WHETHER EXPRESS OR IMPLIED, INCLUDING BUT NOT
--   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
--   PARTICULAR PURPOSE, OR NON-INFRINGEMENT. IN NO EVENT SHALL DIGI-KEY
--   BE LIABLE FOR ANY INCIDENTAL, SPECIAL, INDIRECT OR CONSEQUENTIAL
--   DAMAGES, LOST PROFITS OR LOST DATA, HARM TO YOUR EQUIPMENT, COST OF
--   PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY OR SERVICES, ANY CLAIMS
--   BY THIRD PARTIES (INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF),
--   ANY CLAIMS FOR INDEMNITY OR CONTRIBUTION, OR OTHER SIMILAR COSTS.
--
--   Version History
--   Version 1.0 26.01.2018
--
--------------------------------------------------------------------------------


Library IEEE;
USE ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

Entity SPI_int is
	generic(
		n_bytes_tx : integer := 1;
		n_bytes_rx : integer := 1
	);
	port (
		SCLK : IN std_logic;
		MISO : INOUT std_logic;
		MOSI : IN std_logic;
		SS : IN std_logic;
		reset_n : IN std_logic;
		tx_done : OUT std_logic;
		tx_data : IN std_logic_vector (n_bytes_tx*8-1 downto 0);
		tx_req : OUT std_logic;
		rx_done : OUT std_logic;
		rx_data : OUT std_logic_vector (n_bytes_rx*8-1 downto 0);
		clk : in std_logic
	);
end entity SPI_int;

architecture Behavioral of SPI_int is

type state_type is (init, S1, S2, S3);

constant depth_cnt : integer := integer(ceil(log2(REALMAX(real(n_bytes_rx*8),real(n_bytes_tx*8)))));

signal state : state_type := init;
signal nextstate : state_type;
signal bit_cnt : std_logic_vector (depth_cnt downto 0) := (others => '0');
signal rx_buff : std_logic_vector ((n_bytes_rx)*8-1 downto 0) := (others => '0');
signal tx_buff : std_logic_vector (n_bytes_tx*8-1 downto 0) := (others => '0');
signal tx_req_int : std_logic := '0';
signal tx_req_d : std_logic := '0';
signal en_rx_buff : std_logic := '0';
signal SS_d : std_logic := '1';
signal SS_int : std_logic := '1';
signal SCLK_in_FF : std_logic := '0';
signal SCLK_enh : std_logic := '0';
signal MOSI_in_FF : std_logic := '0';
signal SCLK_int :std_logic := '0';
signal MOSI_int : std_logic := '0';
signal rx_done_int : std_logic := '0';
signal rx_done_d : std_logic_vector (1 downto 0) := "00";
signal rst_rx_buff : std_logic := '0';
signal tx_done_int : std_logic := '0';
signal tx_done_d : std_logic := '0';
signal rst_rx_data : std_logic := '0';

attribute keep_hierarchy : string;
attribute keep_hierarchy of Behavioral : architecture is "yes";

attribute keep : boolean;
attribute keep of SCLK_in_FF : signal is true;
attribute keep of MOSI_in_FF : SIGNAL is true;

begin

rx_done_int <= '1' when (state = S1 and nextstate = init) else '0';
rx_done <= '1' when (rx_done_d (1) = '0' and rx_done_d (0) = '1') else '0';
tx_done_int <= '1' when (state = S2 and nextstate = init) else '0';
tx_done <= '1' when (tx_done_d = '0' and tx_done_int = '1') else '0';
tx_req_int <= '1' when (state = init and nextstate = S2) else '0';
tx_req <= '1' when (tx_req_d <= '0' and tx_req_int = '1') else '0';
MISO <= tx_buff(n_bytes_tx*8 -1) when SS_int = '0' else 'Z';

process(clk,reset_n)
begin
if reset_n = '0' then
	rx_done_d <= "00";
	tx_done_d <= '0';
	tx_req_d <= '0';
elsif rising_edge(clk) then
	--rx_data <= rx_buff;
	rx_done_d (0) <= rx_done_int;
	rx_done_d (1) <= rx_done_d(0);
	tx_done_d <= tx_done_int;
	tx_req_d <= tx_req_int;
end if;
end process;

process(clk,reset_n,rx_done_d,rx_done_int, rst_rx_data)
begin
if reset_n = '0' or rst_rx_data = '1' then
	rx_data <= (others => '0');
elsif rising_edge(clk) then
	if (rx_done_d(0) = '0' and rx_done_int = '1') then
		rx_data <= rx_buff;
	end if;
end if;
end process;

-- SYNCHRONIZER
process(clk)
begin
if rising_edge(clk) then
	SS_d <= SS;
	SS_int <= SS_d;
end if;
end process;
-- MUX RECIRCULATOR

SCLK_in_FF <= SCLK when SS_int = '0' else SCLK_int;
MOSI_in_FF <= MOSI when SS_int = '0' else MOSI_int;

process(clk)
begin
if rising_edge(clk) then
	SCLK_int <= SCLK_in_FF;
	MOSI_int <= MOSI_in_FF;
end if;
end process;


process(reset_n, SCLK_int, SS_int)
begin
	if reset_n = '0' or SS_int = '1' then
		state <= init;
	elsif rising_edge(SCLK_int) then
		state <= nextstate;
	end if;
end process;

process(state,bit_cnt, rx_buff(7 downto 0))
begin
	case state is
		when init =>
			if bit_cnt = std_logic_vector(to_unsigned(8,depth_cnt+1)) then
				case rx_buff(7 downto 0) is
					when "11001100" =>            -- 0xCC	receiving
						nextstate <= S1;
					when "10101010" =>				-- 0xAA  transmitting
						nextstate <= S2;
					when "10011001" =>				-- 0x99  erasing the rx_data
						nextstate <= S3;
					when others =>
						nextstate <= init;
				end case;
			else
				nextstate <= init;
			end if;
		when S1 =>
			if bit_cnt = std_logic_vector(to_unsigned((n_bytes_rx+1)*8,depth_cnt+1)) then
				nextstate <= init;
			else
				nextstate <= S1;
			end if;
		when S2 =>
			if bit_cnt = std_logic_vector(to_unsigned((n_bytes_tx+1)*8,depth_cnt+1)) then
				nextstate <= init;
			else
				nextstate <= S2;
			end if;
		when S3 =>
			if bit_cnt = std_logic_vector(to_unsigned(16,depth_cnt+1)) then
				nextstate <= init;
			else
				nextstate <= S3;
			end if;
		when others =>
			nextstate <= init;
	end case;
end process;

process(state)
begin
	case state is
		when init =>
			en_rx_buff <= '1';
			rst_rx_data <= '0';
		when S1 =>
			en_rx_buff <= '1';
			rst_rx_data <= '0';
		when S2 =>
			en_rx_buff <= '0';
			rst_rx_data <= '0';
		when S3 =>
			en_rx_buff <= '0';
			rst_rx_data <= '1';
		when others =>
			en_rx_buff <= '0';
	end case;
end process;

process(SCLK_int,en_rx_buff,reset_n, SS_int)
begin
	if reset_n = '0' or SS_int = '1' then 
		rx_buff <= (others => '0');
	elsif rising_edge(SCLK_int) then
		if en_rx_buff = '1' then
			rx_buff (0) <= MOSI_int;
			rx_buff ((n_bytes_rx)*8-1 downto 1) <= rx_buff((n_bytes_rx)*8-2 downto 0);
		else
			rx_buff <= rx_buff;
		end if;
	end if;
end process;


process(reset_n, SCLK_int,tx_req_int, SS_int)
begin
	if reset_n = '0' or SS_int = '1' then
		tx_buff <= (others => '0');
	elsif falling_edge(SCLK_int) then
		if tx_req_int = '1' then
			tx_buff <= tx_data;
		elsif state = S2 then
			tx_buff <= tx_buff(n_bytes_tx*8-2 downto 0) & '0';
		end if;
	end if;
end process;

process (reset_n, SCLK_int, SS_int)
begin
	if reset_n = '0' or SS_int = '1' then
		bit_cnt <= (others => '0');
	elsif rising_edge(SCLK_int) then
		bit_cnt <= bit_cnt + 1;
	end if;
end process;
		
end architecture;
