// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module crcFSM
#
(
    parameter I_SIZE = 16,
    parameter TARGET_AW = 13 
)
(
    input clk,
    input syn_rst,
    // From controller
    input read_from_target,
    input [I_SIZE-1:0] I_st, 
    // To CRC
    output crc_feed,
    // mainFSM interface
    input start_compute_run_crc,
    output reg ready,
    output reg fetch_byte_target,
    output [TARGET_AW-1:0] fetch_byte_addr
);

//// Counters
// Counter rec
reg [I_SIZE-1:0] cnt_rec;
reg rst_cnt_rec;
reg inc_cnt_rec;

always@(posedge clk)
if(syn_rst || rst_cnt_rec)
    cnt_rec <= 0;
else if(inc_cnt_rec)
    cnt_rec <= cnt_rec + 1;

wire last_cnt_rec = cnt_rec == I_st-1;

// FSM
reg [2:0] state, nextstate;

localparam IDLE=0,
PRE_FETCH = 1,
FETCH_BYTE = 2,
WAIT_BYTE_TARGET = 3,
FEED_BYTE_CRC = 4;

always@(posedge clk)
if(syn_rst)
    state <= IDLE;
else
    state <= nextstate;


always@(*) begin
    nextstate = state;

    rst_cnt_rec = 0;
    inc_cnt_rec = 0;

    ready = 0;
    fetch_byte_target = 0;

    case(state)
    IDLE: begin
        ready = 1;
        rst_cnt_rec = 1;
        if(start_compute_run_crc) begin
            nextstate = FETCH_BYTE;  
            rst_cnt_rec = 1;
        end
    end
    FETCH_BYTE: begin
        fetch_byte_target = 1;
        nextstate = WAIT_BYTE_TARGET;
    end
    WAIT_BYTE_TARGET: begin
        if(read_from_target) begin
            nextstate = FEED_BYTE_CRC;
        end
    end
    FEED_BYTE_CRC: begin
        inc_cnt_rec = 1;
        if(last_cnt_rec) begin
            nextstate = IDLE;
        end else begin
            nextstate = FETCH_BYTE;
        end
    end
    endcase
end

// Logic for the CRC core
assign crc_feed = inc_cnt_rec;  
assign fetch_byte_addr = cnt_rec[TARGET_AW-1:0];

endmodule
