// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module mainFSM #(
    // parameters
    parameter writeCMD = 8'h22,
    parameter readCMD = 8'h55,
    parameter runCMD = 8'h77,
    parameter pollCMD = 8'hBB,
    parameter read_crcCMD = 8'hEE 
)
(
    input wire clk,
    input wire reset_n,
    input wire [7:0] rx_data_cmd,
    input wire rx_done,
    input wire tx_done,
    input wire tx_req,
    input wire prng_busy,
    input wire prng2_busy,
    input wire [15:0] I_reg,
    input wire [31:0] N_reg,
    input wire [9:0] K_st,
    input wire [1:0] flag,
    output wire T_VALID,
    input wire T_READY,
    output wire reset_target,
    output wire T_START,
    output wire [0:0] en_ctrl_ram_write,
    output wire en_I_reg_counter,
    output reg en_F_shift,
    output wire prng_request,
    output wire prng2_request,
    output reg prng_load_seed,
    output reg prng2_load_seed,
    output reg [2:0] STATUS,
    output reg [9:0] k_counter,
    output wire refresh_fetch_addr,
    output wire refresh_fetch_data,
    output wire refresh_write_data,
    output wire run_start,
    output wire en_N_reg_counter,
    output wire sample_use_base_prng,
    output wire clear_prng_buffer,
    output reg first_run,
    output dyn_clk_start_target_tx,  
    output dyn_clk_start_ctrl_rx,
    output dyn_clk_start_target_rx,
    output dyn_clk_start_config,
    input crc_enable_computation,
    output crc_start_compute_run,
    input crc_core_ready,
    output readcrc_enable
);

// local parameters
localparam INIT = 6'h00;
localparam WAIT_RXDATA = 6'h01;
localparam WRITE_INIT = 6'h02;
localparam WRITE_SAVE = 6'h03;
localparam READ_INIT = 6'h04;
localparam READ_WAIT_DATA = 6'h05;
localparam READ_REQ = 6'h06;
localparam RUN_INIT = 6'h07;
localparam RUN_RESEED_FETCH = 6'h08;
localparam RUN_WAIT_RESEEDING = 6'h09;
localparam RUN_EXTRACT_NFI = 6'h0A;
localparam RUN_EXTRACT_NFI_FINISH = 6'h0B;
localparam RUN_SEND_ROUTINE_INIT = 6'h0C;
localparam RUN_REQUEST_RND = 6'h0D;
localparam RUN_WAIT_PRNG = 6'h0E;
localparam RUN_CRYPTO_INIT = 6'h0F;
localparam RUN_CRYPTO = 6'h10;
localparam RUN_CRYPTO_FINISH = 6'h11;
//	localparam POLL_INIT = 5'h12;
//	localparam POLL_REQ = 5'h13;
localparam RUN_WAIT_D_RESEEDING = 6'h14;
localparam RUN_D_EXTRACT_NFI_FINISH = 6'h15;
localparam RUN_CRYPTO_READY_ACK = 6'h16;
localparam RUN_SEND_ROUTINE_PREFETCH = 6'h17;
localparam REQUEST_PRNG_USE_BASE = 6'h18;
localparam WAIT_PRNG_USE_BASE = 6'h19;
localparam REFRESH_FETCH_ADDRESSES = 6'h1a;
localparam REFRESH_FETCH_DATA = 6'h1b;
localparam REFRESH_WRITE_DATA = 6'h1c;
localparam REFRESH_WAIT_PRNG = 6'h1d;
localparam WAIT_PRNG_CLEAR_FINISHED = 6'h1e;
localparam RUN_REQUEST_RND2 = 6'h1f;
localparam WAIT_DYN_SWITCH_T_TX0 = 6'h20;
localparam WAIT_DYN_SWITCH_T_TX1 = 6'h21;
localparam WAIT_DYN_SWITCH_T_RX0 = 6'h24;
localparam WAIT_DYN_SWITCH_T_RX1 = 6'h25;
localparam WAIT_DYN_SWITCHB_MID0 = 6'h28;
localparam WAIT_DYN_SWITCHB_MID1 = 6'h29;
localparam CRC_COMPUTE_CRC_SWITCH = 6'h2A;
localparam CRC_COMPUTE_WAIT_START_RUN = 6'h2B; 
localparam CRC_COMPUTE_WAIT_RUN = 6'h2C;
localparam READ_CRC_WAIT_FETCH = 6'h2D;
localparam READ_CRC_WAIT_DONE = 6'h2E;

// Randomness bytes
localparam BASE_RND_BYTES = 4;

// state_type
//typedef enum {INIT,WAIT_RXDATA,WRITE_INIT,WRITE_SAVE,
//		READ_INIT,READ_REQ,POLL_INIT,POLL_REQ,
//		RUN_INIT,RUN_RXIST,RUN_CRYPTO} state_type;
reg				[5:0]				state;
reg				[5:0] 			nextstate;


// refresh offset counter
wire en_k_counter, reset_k_counter;
always @(posedge clk) begin
    if (reset_k_counter) k_counter <= 0;
    else if (en_k_counter) k_counter <= k_counter + 1;
end

// state register
always@( posedge clk) begin
    if (~reset_n)
        state <= INIT;
    else
        state <= nextstate;
end

// Temporary counter for intra loop iteration. 
// Used for
// - generation of amount of random bytes 
reg rst_fsm_cnt, inc_fsm_cnt;
reg [7:0] cnt_fsm;

always@(posedge clk)
if(rst_fsm_cnt) begin
    cnt_fsm <= 0;  
end else if(inc_fsm_cnt) begin
    cnt_fsm <= cnt_fsm + 1;
end


// nextstate computation
//always@(state,rx_done,rx_data_cmd,rx_data_f1,rx_data_f2,err_run,tx_done,counter_run) begin
always@(*) begin
    nextstate = state;

    rst_fsm_cnt = 0;
    inc_fsm_cnt = 0;

    case (state)
        // INIT STATE
        INIT : begin
            nextstate = WAIT_DYN_SWITCHB_MID0;
        end
        // WAIT_RXDATA STATE
        WAIT_RXDATA : begin
            if (rx_done)
            case (rx_data_cmd)
                writeCMD : 
                nextstate = WRITE_INIT;
                readCMD :
                nextstate = WAIT_DYN_SWITCH_T_RX0;
                runCMD :
                nextstate = RUN_INIT;
                //pollCMD :
                //	nextstate = POLL_INIT;
                read_crcCMD:
                nextstate = READ_CRC_WAIT_FETCH;
                default :
                nextstate = WAIT_RXDATA;
            endcase
            else
                nextstate = WAIT_RXDATA;
        end

        WRITE_INIT : begin
            nextstate = WRITE_SAVE;
        end

        WRITE_SAVE : begin
            nextstate = WAIT_RXDATA;
        end

        WAIT_DYN_SWITCH_T_RX0: begin
            nextstate = WAIT_DYN_SWITCH_T_RX1;
        end

        WAIT_DYN_SWITCH_T_RX1: begin
            nextstate = READ_INIT;
        end

        READ_INIT : begin
            if (tx_req)
                nextstate = READ_WAIT_DATA;
                // TODO start_target_rx 
            else
                nextstate = READ_INIT;
        end

        READ_WAIT_DATA : begin
            nextstate = READ_REQ;
        end

        READ_REQ : begin
            if (tx_done)
                nextstate = WAIT_DYN_SWITCHB_MID0;
            else
                nextstate = READ_REQ;
        end

        WAIT_DYN_SWITCHB_MID0: begin
            nextstate = WAIT_DYN_SWITCHB_MID1;
        end

        WAIT_DYN_SWITCHB_MID1: begin
            nextstate = WAIT_RXDATA;
        end

        RUN_INIT : begin
            nextstate = RUN_RESEED_FETCH;
        end

        RUN_RESEED_FETCH : begin
            nextstate = RUN_WAIT_D_RESEEDING;
        end

        RUN_WAIT_D_RESEEDING : begin
            nextstate = RUN_WAIT_RESEEDING;
        end

        RUN_WAIT_RESEEDING : begin
            if (prng_busy)
                nextstate = RUN_WAIT_RESEEDING;
            else
                nextstate = RUN_EXTRACT_NFI;
        end

        READ_CRC_WAIT_FETCH: begin
            if(tx_req) begin
                nextstate = READ_CRC_WAIT_DONE;
            end
        end

        READ_CRC_WAIT_DONE: begin
            if(tx_done) begin
                nextstate = WAIT_DYN_SWITCHB_MID0;
            end
        end

        /// When looping over executions, we arrive here !
        RUN_EXTRACT_NFI : begin
            nextstate = RUN_D_EXTRACT_NFI_FINISH;
            // TODO start_start_config 
        end

        //////////
        RUN_D_EXTRACT_NFI_FINISH : begin
            nextstate = RUN_EXTRACT_NFI_FINISH;
        end
        RUN_EXTRACT_NFI_FINISH : begin
            //nextstate <= RUN_SEND_ROUTINE_PREFETCH;//RUN_SEND_ROUTINE_INIT;
            nextstate = REFRESH_FETCH_ADDRESSES;
        end
        /// refresh RAM
        REFRESH_FETCH_ADDRESSES: begin
            if (k_counter == K_st) begin
                nextstate = REQUEST_PRNG_USE_BASE;
                rst_fsm_cnt = 1;
            end else begin
                nextstate = REFRESH_FETCH_DATA;
            end
        end
        REFRESH_FETCH_DATA: begin
            nextstate = REFRESH_WRITE_DATA;
        end
        REFRESH_WRITE_DATA: begin
            nextstate = REFRESH_WAIT_PRNG;
        end
        REFRESH_WAIT_PRNG: begin
            if (prng_busy) nextstate = REFRESH_WAIT_PRNG;
            else nextstate = REFRESH_FETCH_ADDRESSES;
        end
        ////
        /// use_base,
        /// in REQUEST_PRNG_USE_BASE: sample prng's output (that's already fresh)
        /// & set request_prng_output
        /// - request_prng_use_base state -> 1 cycle, set prng_request to 1
        /// if prng_ready, sample prng output in reg for use_base
        /// in WAIT_PRNG_USE_BASE: 
        /// - wait prng ready state use_base -> loop while !prng_ready,
        /// - continue
        ////

        REQUEST_PRNG_USE_BASE: begin
            nextstate = WAIT_PRNG_USE_BASE;
        end
        WAIT_PRNG_USE_BASE: begin
            if (prng_busy) begin 
                nextstate = WAIT_PRNG_USE_BASE;
            end else begin
                if (cnt_fsm == BASE_RND_BYTES-1) begin
                    nextstate = WAIT_DYN_SWITCH_T_TX0;
                end else begin
                    inc_fsm_cnt = 1;
                    nextstate = REQUEST_PRNG_USE_BASE;
                end
            end
        end
        //////////
        // STATES to properly switch between clocks
        WAIT_DYN_SWITCH_T_TX0: begin
            nextstate = WAIT_DYN_SWITCH_T_TX1;
        end
        WAIT_DYN_SWITCH_T_TX1: begin
            nextstate = RUN_SEND_ROUTINE_PREFETCH;
        end
        /////////
        RUN_SEND_ROUTINE_PREFETCH : begin
            nextstate = RUN_SEND_ROUTINE_INIT;
            // TODO start_target_tx
        end

        RUN_SEND_ROUTINE_INIT : begin
            if (flag == 1)
                nextstate = RUN_REQUEST_RND;
            else if (flag == 2) 
                nextstate = RUN_REQUEST_RND2;
            else begin
                if (I_reg == 0)
                    nextstate = RUN_CRYPTO_INIT;
                else
                    nextstate = RUN_SEND_ROUTINE_INIT;
            end

            /*
            if (I_reg == 0)
                    nextstate = RUN_CRYPTO_INIT;
            else
                    if (flag == 1'b1)
                            nextstate = RUN_REQUEST_RND;
                    else
                            nextstate = RUN_SEND_ROUTINE_INIT;
                        */
        end

        RUN_REQUEST_RND: begin
            if (prng_busy)
                nextstate = RUN_REQUEST_RND;
            else
            if (I_reg == 0)
                nextstate = RUN_CRYPTO_INIT;
            else	
                nextstate = RUN_SEND_ROUTINE_PREFETCH;//RUN_SEND_ROUTINE_INIT;//RUN_WAIT_PRNG;
        end

        RUN_REQUEST_RND2: begin
            if (prng2_busy)
                nextstate = RUN_REQUEST_RND2;
            else
            if(I_reg == 0)
                nextstate = RUN_CRYPTO_INIT;
            else
                nextstate = RUN_SEND_ROUTINE_PREFETCH;
        end

        //			RUN_WAIT_PRNG : begin
        //				nextstate <= RUN_SEND_ROUTINE_INIT;
        //			end

        RUN_CRYPTO_INIT : begin
            nextstate = RUN_CRYPTO;
            // here, T_START is aserted
        end

        RUN_CRYPTO : begin
            if (T_READY)
                nextstate = RUN_CRYPTO_READY_ACK;
            else
                nextstate = RUN_CRYPTO;
        end

        RUN_CRYPTO_READY_ACK : begin
            nextstate = WAIT_PRNG_CLEAR_FINISHED;
        end

        WAIT_PRNG_CLEAR_FINISHED: begin
            if (prng_busy) begin
                nextstate = WAIT_PRNG_CLEAR_FINISHED;
            end else begin
                if(crc_enable_computation) begin
                    nextstate = CRC_COMPUTE_CRC_SWITCH;
                end else begin
                  nextstate = RUN_CRYPTO_FINISH;
                end
            end
        end
        CRC_COMPUTE_CRC_SWITCH: begin
            nextstate = CRC_COMPUTE_WAIT_START_RUN;
        end
        CRC_COMPUTE_WAIT_START_RUN: begin
            if(crc_core_ready) begin
                nextstate = CRC_COMPUTE_WAIT_RUN;
            end
        end
        CRC_COMPUTE_WAIT_RUN: begin
            if(crc_core_ready) begin
                nextstate = RUN_CRYPTO_FINISH;
            end
        end
        RUN_CRYPTO_FINISH : begin
            if (N_reg == 0)
                nextstate = WAIT_DYN_SWITCHB_MID0;
                // TODO start_target RX
            else
                nextstate = RUN_EXTRACT_NFI;
        end

        //			POLL_INIT : begin
        //				if (tx_req)
        //					nextstate <= POLL_REQ;
        //				else
        //					nextstate <= POLL_INIT;
        //			end
        //			
        //			POLL_REQ : begin
        //				if (tx_done)
        //					nextstate <= WAIT_RXDATA;
        //				else
        //					nextstate <= POLL_REQ;
        //			end

        default : begin
            nextstate = INIT;
        end
    endcase
end

// status
always@(*)begin
    case(state)
        WRITE_INIT: STATUS <= 3'b000;
        WRITE_SAVE: STATUS <= 3'b000;
        RUN_INIT : STATUS <= 3'b001;
        RUN_RESEED_FETCH : STATUS <= 3'b001;
        RUN_WAIT_RESEEDING : STATUS <= 3'b001;
        RUN_EXTRACT_NFI : STATUS <= 3'b001;
        RUN_EXTRACT_NFI_FINISH : STATUS <= 3'b001;
        RUN_SEND_ROUTINE_PREFETCH : STATUS <= 3'b100;
        RUN_SEND_ROUTINE_INIT : STATUS <= 3'b100;
        RUN_REQUEST_RND : STATUS <= 3'b100;
        RUN_WAIT_PRNG : STATUS <= 3'b100;
        RUN_CRYPTO_INIT : STATUS <= 3'b001;
        RUN_CRYPTO : STATUS <= 3'b001;
        RUN_CRYPTO_FINISH : STATUS <= 3'b001;	
        RUN_CRYPTO_READY_ACK : STATUS <= 3'b001;
        RUN_WAIT_D_RESEEDING : STATUS <= 3'b001;
        RUN_D_EXTRACT_NFI_FINISH : STATUS <= 3'b001;
        //			POLL_INIT : STATUS <= 3'b010;
        //			POLL_REQ : STATUS <= 3'b010;
        READ_INIT : STATUS <= 3'b011;
        READ_WAIT_DATA : STATUS <= 3'b011;
        READ_REQ : STATUS <= 3'b011;
        WAIT_RXDATA : STATUS <= 3'b101;
        WAIT_PRNG_CLEAR_FINISHED: STATUS <= 3'b001;
        default : STATUS <= 3'b111;
    endcase 
end

// reset the target fpga on the INIT (to be exteneded to other states)
assign reset_target = ((state == RUN_EXTRACT_NFI)) ? 1'b0 : 1'b1;
// T_START lasts 1 clock cycle
assign T_START = (state == RUN_CRYPTO_INIT) ? 1'b1 : 1'b0;
// en_ctrl_ram_write is '1' only if a write is requested
assign en_ctrl_ram_write = (state == WRITE_INIT) ? 1 : 0;
// issue a request for data
//assign tx_req = (state == READ_INIT) ? 1'b1 : 1'b0;
// advance of one word in pushing data in the target FPGA
wire en_I_reg_counter_int = (nextstate == RUN_SEND_ROUTINE_INIT) ? 1'b1 : 1'b0;
assign run_start = (state == RUN_EXTRACT_NFI);
assign en_I_reg_counter = en_I_reg_counter_int;
assign en_N_reg_counter = (state == RUN_CRYPTO_READY_ACK);
assign clear_prng_buffer = (state == RUN_CRYPTO_READY_ACK);
always@(posedge clk)begin
    en_F_shift <= en_I_reg_counter;
end

//assign en_F_shift = en_I_reg_counter;

wire t_valid_DD = ((state == WAIT_DYN_SWITCH_T_RX1) && (nextstate == READ_INIT) ) ? 1'b1 : 1'b0;
reg t_valid_D;
always@(posedge clk)begin
    if (~reset_n)
        t_valid_D <= 0;
    else
        t_valid_D <= t_valid_DD;
end

assign T_VALID = en_I_reg_counter | t_valid_D;
// prng_request (lasts 1 clock cycle)
//assign prng_request = ((state == RUN_SEND_ROUTINE_INIT) &&
//								(nextstate == RUN_REQUEST_RND)) ? 1'b1 : 1'b0;
assign prng_request = ((state == RUN_SEND_ROUTINE_INIT) && (flag==1)) | (state == REQUEST_PRNG_USE_BASE) | (state == REFRESH_WRITE_DATA);
assign sample_use_base_prng = (state == REQUEST_PRNG_USE_BASE);

assign prng2_request = (state == RUN_SEND_ROUTINE_INIT) && (flag == 2);

// prng_load_seed (lasts 1 clock cycle)
//wire prng_load_seed_int = ((state == RUN_RESEED_FETCH) &
//								(nextstate == RUN_WAIT_RESEEDING)) ? 1'b1 : 1'b0;

//wire prng_load_seed_int = (state == RUN_WAIT_D_RESEEDING) ? 1'b1 : 1'b0;								
wire prng_load_seed_int = (state == RUN_RESEED_FETCH) ? 1'b1 : 1'b0;								
always@(posedge clk)begin
    if (~reset_n) begin
        prng_load_seed <= 0;
        prng2_load_seed <= 0;
    end else begin
        prng_load_seed <= prng_load_seed_int;
        prng2_load_seed <= prng_load_seed_int;
    end
end


always@(posedge clk) begin
    if (~reset_n)
        first_run <= 1;
    else if (state == RUN_INIT)
        first_run <= 1;
    else if ((state== RUN_CRYPTO_INIT) && (nextstate == RUN_CRYPTO))
        first_run <= 0;
end

assign reset_k_counter = (state == RUN_EXTRACT_NFI);
assign en_k_counter = (state == REFRESH_WRITE_DATA);
assign refresh_fetch_addr = (state == REFRESH_FETCH_ADDRESSES);
assign refresh_fetch_data = (state == REFRESH_FETCH_DATA);
assign refresh_write_data = (state == REFRESH_WRITE_DATA);

// Dynamic clock reconfiguration
assign dyn_clk_start_target_tx = (nextstate == WAIT_DYN_SWITCH_T_TX0) || (state == CRC_COMPUTE_WAIT_START_RUN);
assign dyn_clk_start_ctrl_rx = nextstate == WAIT_DYN_SWITCHB_MID0;
assign dyn_clk_start_target_rx = nextstate == WAIT_DYN_SWITCH_T_RX0;
assign dyn_clk_start_config = (state == RUN_CRYPTO_INIT);

// TODO crc_start compute
assign crc_start_compute_run = state == CRC_COMPUTE_WAIT_START_RUN;

assign readcrc_enable = (state == READ_CRC_WAIT_FETCH) || (state == READ_CRC_WAIT_DONE);

endmodule
