// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module ucl_cg_DCM_CLKGEN_controller
(
    input clk,
    input rst,
    // DATA
    input [7:0] D_data,
    input [7:0] M_data,
    // SPI master
    output PROGCLK,
    output reg  PROGEN,
    output PROGDATA,
    input PROGDONE,
    input LOCKED,
    // to main FSM control
    output reg PROG_CORE_READY,
    input PROG_CORE_START
);

// LOCALPARAM for minimal gap
localparam MIN_GAP_LOAD = 2;
localparam MIN_GAP_GO = 1;

// Buffer for the data
reg [20:0] buf_DM;
wire valid_start = PROG_CORE_READY && PROG_CORE_START;
reg shift_DM_buffer;

always@(posedge clk)
if(valid_start) begin
    buf_DM <= {1'b1,M_data,2'b11,D_data,2'b01};
end else if(shift_DM_buffer) begin
    buf_DM <= {1'b0,buf_DM[20:1]};
end

// Fsm counter
reg [4:0] fsm_cnt;
reg rst_fsm_cnt, inc_fsm_cnt;

always@(posedge clk) 
if(rst || rst_fsm_cnt) begin
    fsm_cnt <= 0;
end else if(inc_fsm_cnt) begin
    fsm_cnt <= fsm_cnt + 1;
end

wire last_load_cmd_cycle = fsm_cnt == 1;
wire last_load_data_cycle = fsm_cnt == 7;
wire last_gap_load_cycle = fsm_cnt == (MIN_GAP_LOAD-1);
wire last_gap_go_cycle = fsm_cnt == (MIN_GAP_GO-1);

// FSM
reg [4:0] state, nextstate;

localparam IDLE = 0,
LOAD_D_cmd = 1,
LOAD_D_data = 2,
GAP_LOAD = 3,
LOAD_M_cmd = 4,
LOAD_M_data = 5,
GAP_GO = 6,
GO_cmd = 7,
WAIT_FINISH = 8;

always@(posedge clk)
if(rst) begin
    state <= WAIT_FINISH;
end else begin
    state <= nextstate;
end

// FSM logic
always@(*) begin
    nextstate = state;

    rst_fsm_cnt = 0;    
    shift_DM_buffer = 0;
    inc_fsm_cnt = 0;

    PROG_CORE_READY = 0;
    
    PROGEN = 0;

    case(state)
        IDLE: begin
            PROG_CORE_READY = 1;
            if(PROG_CORE_START) begin
                rst_fsm_cnt = 1;
                nextstate = LOAD_D_cmd;     
            end
        end
        LOAD_D_cmd: begin
            PROGEN = 1;
            inc_fsm_cnt = 1;
            shift_DM_buffer = 1;
            if(last_load_cmd_cycle) begin
                nextstate = LOAD_D_data;
                rst_fsm_cnt = 1;
            end
        end
        LOAD_D_data: begin
            PROGEN = 1;
            inc_fsm_cnt = 1;
            shift_DM_buffer = 1;
            if(last_load_data_cycle) begin
                nextstate = GAP_LOAD;
                rst_fsm_cnt = 1;
            end
        end
        GAP_LOAD: begin
            inc_fsm_cnt = 1;
            if(last_gap_load_cycle) begin
                nextstate = LOAD_M_cmd;
                rst_fsm_cnt = 1;
            end
        end
        LOAD_M_cmd: begin
            PROGEN = 1;
            inc_fsm_cnt = 1;
            shift_DM_buffer = 1;
            if(last_load_cmd_cycle) begin
                nextstate = LOAD_M_data;
                rst_fsm_cnt = 1;
            end
        end
        LOAD_M_data: begin
            PROGEN = 1;
            inc_fsm_cnt = 1;
            shift_DM_buffer = 1;
            if(last_load_data_cycle) begin
                nextstate = GAP_GO;
                rst_fsm_cnt = 1;
            end
        end
        GAP_GO: begin
            inc_fsm_cnt = 1;
            if(last_gap_go_cycle) begin
                nextstate = GO_cmd;
                rst_fsm_cnt = 1;
            end
        end
        GO_cmd: begin
            PROGEN = 1;
            shift_DM_buffer = 1;
            nextstate = WAIT_FINISH;
        end
        WAIT_FINISH: begin
            if(PROGDONE && LOCKED) begin
                nextstate = IDLE;
            end
        end
    endcase
end

// Assign outputs
assign PROGCLK = clk;
assign PROGDATA = buf_DM[0];

endmodule
