// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module ucl_cg_top_control(
    // INPUT CLOCK
    input wire C_CLK,
    input wire C_CLK_EXT,
    input wire C_RESET_B,
    // SPI SIGNALS 
    input wire SCLK,
    input wire SS,
    input wire MOSI,
    output wire MISO,
    // GPPIO
    inout [15:0] C_GPIO,
    // LEDS (for debug)
    output wire [7:0] LED_STATUS,
    // TARGET BUS (APB-LOOKALIKE)
    // OUTPUT CLOCK FOR TARGET
    output wire T_CLK,
    // ADDRESS BUS 8BIT
    output wire [TARGET_AW-1:0] T_ADDR,
    // WRITE DATA BUS 8BIT (CONTROL->TARGET)
    output wire [7:0] T_WDATA,
    // READ DATA BUS 8BIT (TARGET->CONTROL)
    input wire [7:0] T_RDATA,
    // '1' TO ENABLE CRYPTOGRAPHIC OPERATION ON TARGET
    output wire T_START,
    // '1' TO WRITE ON INPUT MEMORY OF TARGET, '0' IN READ-MODE
    output reg T_WRITE,
    // '1' TO COMMUNICATE A VALID MESSAGE ON THE BUS
    output reg T_VALID,
    // '1' WHEN TARGET OPERATION IS FINISHED
    input wire T_READY,
    // RESET TARGET
    output wire T_RSTB,
    // T_REAL_EXEC
    output wire T_REAL_EXEC
);

localparam TARGET_AW = 13;
localparam INNER_AW = 16;
localparam TARGET_RAM_SIZE = 1024; // bytes (=words)
localparam I_SIZE = 16; // bits
localparam N_SIZE = 32; // bits
localparam B_SIZE = 16; // bits
localparam K_SIZE = 16; // bits
localparam INDIV_F_SIZE = 2; // bits
localparam F_SIZE = TARGET_RAM_SIZE*INDIV_F_SIZE; // bits
localparam MISC_SIZE = 8; // bits
localparam CRC_BYTES_SIZE = 16;
localparam NSTATES_SIZE = 8; //bits

localparam I_OFFSET = 'h0000;
localparam SEED_OFFSET = 'h0002;
localparam N_OFFSET = 'h0006;
localparam MISC_OFFSET = 'h000a;
localparam B_OFFSET = 'h000b;
localparam K_OFFSET = 'h000d;
localparam F_OFFSET = 'h0018;
localparam SEED2_OFFSET = 'h0010;
localparam CRC_BYTES_OFFSET = 'h0014;
localparam NSTATES_OFFSET = 'h0016;

localparam BASE_RND_BYTES = 4;

// Generation parameters
localparam SPI_RX_BYTE = 4;
localparam SPI_TX_BYTE = 1;
localparam writeCMD = 8'h22;
localparam csrCMD = 8'h33;
localparam readCMD = 8'h55;
localparam runCMD = 8'h77;
localparam pollCMD = 8'hBB;
localparam resetCMD = 8'hAA;
localparam rfrsh0CMD = 8'hCC;
localparam rfrsh1CMD = 8'hDD;
localparam read_crcCMD = 8'hEE;

// Buffered input wire clock
localparam BUFFERED_IN_CLK = 1;

wire clk;

// Inverted output clock
localparam INV_OUT_CLK = 0;


// Internal signals
reg    [I_SIZE-1:0]    I_reg;
wire    [I_SIZE-1:0]    I_st;
wire    [B_SIZE-1:0]    B_st;
wire    [K_SIZE-1:0]    K_st;
reg    [N_SIZE-1:0]    N_reg;
wire    [N_SIZE-1:0]    N_st;
reg    [F_SIZE-1:0]    F_reg;
wire    [F_SIZE-1:0]    F_st;
wire [NSTATES_SIZE-1:0] NSTATES_st;

wire [INDIV_F_SIZE-1:0] F_used = F_reg[F_SIZE-1 -: INDIV_F_SIZE];
//wire [INDIV_F_SIZE-1:0] F_used = F_st[F_SIZE-1 -: INDIV_F_SIZE];

// misc flags: misc_st[0] reset target at every execution
wire    [MISC_SIZE-1:0]    misc_st;
reg    T_READY_REG;
reg    [7:0]    T_RDATA_REG;
reg    [TARGET_AW-1:0]    addr_reg, addr_reg_d;
reg    soft_reset_b;
wire csr_done;
// offset the address with the base pointer
reg [8*BASE_RND_BYTES-1:0] rnd_use_base;
wire reset_target_fsm;

wire en_I_reg_counter;
wire en_N_reg_counter;

// CRC read data
reg [7:0] crc_byte_read;

// POLL AUTH signal
reg    poll_eng_auth;
reg [7:0] prev_state;

// CTRL_RAM signals
wire    [7:0]    ram_data_out;
wire     [0:0]    en_ctrl_ram_write;
wire     [7:0]     ctrl_RAM_data;
reg     [TARGET_AW-1:0]     ctrl_RAM_address0;
reg [TARGET_AW-1:0]     ctrl_RAM_address1;

// PRNG1 signals
wire    prng_request;
wire    prng_load_seed;
wire    prng_busy;
wire    [7:0]    prng_out;
wire    [31:0]    prng_seed;

// PRNG2 signals
wire prng2_request;
wire prng2_load_seed;
wire prng2_busy;
wire [7:0] prng2_out;
wire [31:0] prng2_seed;

// FSM signals
wire    run_start;
wire    [2:0]    STATUS;
wire    first_run;
wire    T_VALID_int;
wire sample_use_base_prng;
wire [9:0] k_counter;
wire refresh_fetch_addr, refresh_fetch_data, refresh_write_data;
wire clear_prng_buffer;

wire dyn_clk_start_target_tx;  
wire dyn_clk_start_ctrl_rx;
wire dyn_clk_start_target_rx;
wire dyn_clk_start_config;

wire crc_enable_computation;
wire crc_start_compute_run;
wire crc_core_ready;

wire readcrc_enable;


// SPI interface
// SPI signals
wire    tx_done;
wire    [8*SPI_TX_BYTE-1:0]    tx_data;
wire    tx_req;
wire    rx_done;
wire    [8*SPI_RX_BYTE-1:0]    rx_data;
wire    [7:0]    rx_data_cmd;
wire    [15:0]    rx_data_f1;
wire    [7:0]    rx_data_f2;

// refresh RAM signals
wire [9:0] refresh_ram_address0;
wire [9:0] refresh_ram_address1;
reg [7:0] refresh_fetched_addr_lsb0, refresh_fetched_addr_msb0, refresh_fetched_addr_lsb1, refresh_fetched_addr_msb1;

SPI_int 
#(
    .n_bytes_tx(SPI_TX_BYTE),
    .n_bytes_rx(SPI_RX_BYTE)
) 
spi(
    .SCLK(SCLK),
    .MISO(MISO),
    .MOSI(MOSI),
    .SS(SS),
    .reset_n(soft_reset_b),
    .tx_done(tx_done),
    .tx_data(tx_data),
    .tx_req(tx_req),
    .rx_done(rx_done),
    .rx_data(rx_data),
    .clk(clk)
);

assign rx_data_cmd = rx_data[24 +: 8];
assign rx_data_f1 = rx_data[8 +: 16];
assign rx_data_f2 = rx_data[0 +: 8];
assign tx_data = readcrc_enable ? crc_byte_read : poll_eng_auth ? prev_state : T_RDATA_REG;

always@(posedge clk)begin
    if (rx_data_cmd == resetCMD)
        soft_reset_b <= 1'b0;
    else
        soft_reset_b <= C_RESET_B;
end



// Main FSM
mainFSM 
#(
    .writeCMD(writeCMD),
    .readCMD(readCMD),
    .runCMD(runCMD),
    .pollCMD(pollCMD),
    .read_crcCMD(read_crcCMD)
)
mfsm(
    .clk(clk),
    .reset_n(soft_reset_b),
    .rx_data_cmd(rx_data_cmd),
    .rx_done(rx_done),
    .tx_done(tx_done),
    .tx_req(tx_req),
    .prng_busy(prng_busy),
    .prng2_busy(prng2_busy),
    .I_reg(I_reg),
    .N_reg(N_reg),
    .K_st(K_st),
    .flag(F_used),
    .T_VALID(T_VALID_int),
    .T_READY(T_READY_REG),
    .reset_target(reset_target_fsm),
    .T_START(T_START),
    .en_ctrl_ram_write(en_ctrl_ram_write),
    .en_I_reg_counter(en_I_reg_counter),
    .en_F_shift(en_F_shift),
    .prng_request(prng_request),
    .prng2_request(prng2_request),
    .prng_load_seed(prng_load_seed),
    .prng2_load_seed(prng2_load_seed),
    .STATUS(STATUS),
    .k_counter(k_counter),
    .refresh_fetch_addr(refresh_fetch_addr),
    .refresh_fetch_data(refresh_fetch_data),
    .refresh_write_data(refresh_write_data),
    .run_start(run_start),
    .en_N_reg_counter(en_N_reg_counter),
    .sample_use_base_prng(sample_use_base_prng),
    .clear_prng_buffer(clear_prng_buffer),
    .first_run(first_run),
    .dyn_clk_start_target_tx(dyn_clk_start_target_tx),
    .dyn_clk_start_ctrl_rx(dyn_clk_start_ctrl_rx),
    .dyn_clk_start_target_rx(dyn_clk_start_target_rx),
    .dyn_clk_start_config(dyn_clk_start_config),
    .crc_enable_computation(crc_enable_computation),
    .crc_start_compute_run(crc_start_compute_run),
    .crc_core_ready(crc_core_ready),
    .readcrc_enable(readcrc_enable)
);

// CONTROL FPGA RAM
/*
CTRL_RAM ram(
                                 .clka(clk),
                                 .rsta(~soft_reset_b),
                                 .wea(en_ctrl_ram_write),
                                 .addra(ctrl_RAM_address0),
                                 .dina(ctrl_RAM_data),
                                 .douta(ram_data_out));
                                 */
reg [7:0] main_ram_state [0:8192-1];
always @(posedge clk) begin
    if (en_ctrl_ram_write) main_ram_state[ctrl_RAM_address0] <= ctrl_RAM_data;
    else if (refresh_write_data) main_ram_state[ctrl_RAM_address0] <= ram_dout ^ prng_out;
    if (refresh_write_data) main_ram_state[ctrl_RAM_address1] <= ram_dout2 ^ prng_out;
end
reg [7:0] ram_dout, ram_dout2;
always @(posedge clk) begin
    ram_dout <= main_ram_state[ctrl_RAM_address0];
    ram_dout2 <= main_ram_state[ctrl_RAM_address1];
end
assign ram_data_out = ram_dout;

// TODO: here fix use base with mod
localparam BITS_MODULO = 3;
reg [BITS_MODULO-1:0] ptr_state_index;
wire [TARGET_AW-1:0] ptr_state_index_bytes = ptr_state_index * B_st;

always@(*)begin
    if (STATUS == 3'b100) ctrl_RAM_address0 <= addr_reg + ptr_state_index_bytes;
    else if (en_ctrl_ram_write) ctrl_RAM_address0 <= rx_data_f1[0 +: TARGET_AW];
    else if (refresh_fetch_data | refresh_write_data) ctrl_RAM_address0 <= {refresh_fetched_addr_msb0[0 +: (TARGET_AW-8)], refresh_fetched_addr_lsb0};
    else ctrl_RAM_address0 <= 0;
end
always@(*)begin
    if (refresh_fetch_data | refresh_write_data) ctrl_RAM_address1 <= {refresh_fetched_addr_msb1[0 +: (TARGET_AW-8)], refresh_fetched_addr_lsb1};
    else ctrl_RAM_address1 <= 0;
end
//assign ctrl_RAM_address1 = {refresh_fetched_addr_msb1, refresh_fetched_addr_lsb1};
assign ctrl_RAM_data = rx_data_f2;

// Refresh RAM
reg [7:0] refresh_ram_state_lsb [0:1024-1];
reg [7:0] refresh_ram_state_msb [0:1024-1];
assign refresh_ram_address0 = (en_refresh_ram_write_lsb | en_refresh_ram_write_msb) ? rx_data_f1[0 +: 10] : 2*k_counter;
assign refresh_ram_address1 = 2*k_counter+1;
assign en_refresh_ram_write_lsb = rx_done && (rx_data_cmd == rfrsh0CMD);
assign en_refresh_ram_write_msb = rx_done && (rx_data_cmd == rfrsh1CMD);
/// Refresh RAM Port 0
always @(posedge clk) begin
    if (en_refresh_ram_write_lsb) refresh_ram_state_lsb[refresh_ram_address0] <= rx_data_f2;
end
always @(posedge clk) begin
    if (en_refresh_ram_write_msb) refresh_ram_state_msb[refresh_ram_address0] <= rx_data_f2;
end
always @(posedge clk) begin
    refresh_fetched_addr_lsb0 <= refresh_ram_state_lsb[refresh_ram_address0];
end
always @(posedge clk) begin
    refresh_fetched_addr_msb0 <= refresh_ram_state_msb[refresh_ram_address0];
end
/// Refresh RAM Port 1
always @(posedge clk) begin
    refresh_fetched_addr_lsb1 <= refresh_ram_state_lsb[refresh_ram_address1];
end
always @(posedge clk) begin
    refresh_fetched_addr_msb1 <= refresh_ram_state_msb[refresh_ram_address1];
end


assign csr_done = rx_done && (rx_data_cmd == csrCMD);


// PRNG1 WRAPPER
prng_wrapper
prng(
    .clk(clk),
    .reset_n(soft_reset_b),
    .seed(prng_seed),
    .prng_request(prng_request),
    .prng_load_seed(prng_load_seed),
    .clear_prng_buffer(clear_prng_buffer),
    .dout(prng_out),
    .prng_busy(prng_busy)
);

csr #(
    .WORDS(4),
    .AW(INNER_AW),
    .OFFSET(SEED_OFFSET)
)
prng_seed_csr (
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(prng_seed)
);

// PRNG2 WRAPPER
prng_wrapper
prng2(
    .clk(clk),
    .reset_n(soft_reset_b),
    .seed(prng2_seed),
    .prng_request(prng2_request),
    .prng_load_seed(prng2_load_seed),
    .clear_prng_buffer(clear_prng_buffer),
    .dout(prng2_out),
    .prng_busy(prng2_busy)
);

csr #(
    .WORDS(4),
    .AW(INNER_AW),
    .OFFSET(SEED2_OFFSET)
)
prng2_seed_csr(
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(prng2_seed)
);


// Controller for the dynamic clock generator
wire clk_ctrl, clk_target;
wire clk_debug;


clock_unit clk_u(
    .CLK_OSC(C_CLK),
    .CLK_EXT(C_CLK_EXT),
    .CLK_SW(misc_st[2]),
    .dyn_clk_start_target_tx(dyn_clk_start_target_tx),
    .dyn_clk_start_ctrl_rx(dyn_clk_start_ctrl_rx),
    .dyn_clk_start_target_rx(dyn_clk_start_target_rx),
    .dyn_clk_start_config(dyn_clk_start_config),
    .CLK_CTRL(clk_ctrl),
    .CLK_TARGET(clk_target),
    .CLK_DEBUG(clk_debug)
);

assign clk = clk_ctrl;

wire debug_clk_ctrl;
wire debug_clk_target;
ODDR2 CB_ctrl(.D0(1'b0),.D1(1'b1),.C0(~clk_debug),.C1(clk_debug),.Q(debug_clk_ctrl),.CE(1'b1),.R(1'b0));
ODDR2 CB_target(.D0(1'b0),.D1(1'b1),.C0(~clk_target),.C1(clk_target),.Q(debug_clk_target),.CE(1'b1),.R(1'b0));


////// number of cryptographic operation (e.g. encryption)
csr #(
    .WORDS(N_SIZE/8),
    .AW(INNER_AW),
    .OFFSET(N_OFFSET),
    .BYTESWAP(1)
)
N_csr (
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(N_st)
);
always@(posedge clk)begin
    if (~soft_reset_b)
        N_reg <= 0;
    else if (run_start && first_run)
        N_reg <= (N_st);
    else if (en_N_reg_counter) //(T_READY_REG) // FIXME do this based on FSM state
        N_reg <= N_reg - 1;
end

assign T_REAL_EXEC = (N_reg <= N_st);

// F_reg[i] represents if the byte should be deterministic
// or from the PRNG
csr #(
    .WORDS(F_SIZE/(8)),
    .AW(INNER_AW),
    .OFFSET(F_OFFSET),
    .BYTESWAP(1)
)
F_csr (
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(F_st)
);
always@(posedge clk)begin
    if (~soft_reset_b)
        F_reg <= 0;
    else if (run_start)
        F_reg <= F_st;
    else if (en_F_shift)
        F_reg <= {F_reg[F_SIZE-INDIV_F_SIZE-1:0],F_reg[F_SIZE-1 -: INDIV_F_SIZE]};
end

// I_reg has the number of bytes to send to the target
csr #(
    .WORDS(I_SIZE/8),
    .AW(INNER_AW),
    .OFFSET(I_OFFSET),
    .BYTESWAP(1)
)
I_csr (
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(I_st)
);
always@(posedge clk)begin
    if (~soft_reset_b)
        I_reg <= 0;
    else if (run_start)
        I_reg <= I_st;
    else if (en_I_reg_counter)
        I_reg <= I_reg - 1;
end

// MISC csr
//
// misc_st[0] = reset
// misc_st[1] = CRC computation
// misc_st[2] = clk switch
csr #(
    .WORDS(MISC_SIZE/8),
    .AW(INNER_AW),
    .OFFSET(MISC_OFFSET),
    .BYTESWAP(1)
)
misc_csr (
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(misc_st)
);

// B_st has the number of bytes to send to the target
csr #(
    .WORDS(B_SIZE/8),
    .AW(INNER_AW),
    .OFFSET(B_OFFSET),
    .BYTESWAP(1)
)
B_csr (
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(B_st)
);

// K_st has the number refresh pairs
csr #(
    .WORDS(K_SIZE/8),
    .AW(INNER_AW),
    .OFFSET(K_OFFSET),
    .BYTESWAP(1)
)
K_csr (
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(K_st)
);

// NSTATES_st has the module to apply for multiple configuration
csr #(
    .WORDS(NSTATES_SIZE / 8),
    .AW(INNER_AW),
    .OFFSET(NSTATES_OFFSET),
    .BYTESWAP(1)
)
Bmod_csr (
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(NSTATES_st)
);

// register for the address of the target FPGA to write
always@(posedge clk)begin
    if ((~soft_reset_b) || run_start)
        addr_reg <= 0;
    else if (en_I_reg_counter)
        addr_reg <= addr_reg + 1;
end
always@(posedge clk)begin
    if (~soft_reset_b)
        addr_reg_d <= 0;
    else
        addr_reg_d <= addr_reg;    
end


// input wire register for the incoming signals from the target FPGA
// sampled (to gain on timing)
always@(posedge clk) begin
    T_READY_REG <= T_READY;
end

always@(posedge clk) begin
    if (~soft_reset_b)
        T_RDATA_REG <= 0;
    else if(en_read)
        T_RDATA_REG <= T_RDATA;
end

wire crc_fetch_byte_sig;
always@(posedge clk)begin
    T_VALID <= T_VALID_int || crc_fetch_byte_sig;
end

reg en_read;
always@(posedge clk)begin
    en_read <= T_VALID & ~T_WRITE;
end

////// CRC core
// B_st has the number of bytes to send to the target
wire [CRC_BYTES_SIZE-1:0] CRC_BYTES_st; 
csr #(
    .WORDS(CRC_BYTES_SIZE/8),
    .AW(INNER_AW),
    .OFFSET(CRC_BYTES_OFFSET),
    .BYTESWAP(1)
)
CRCB_csr (
    .clk(clk),
    .reset_n(soft_reset_b),
    .write(csr_done),
    .address(rx_data_f1),
    .ctrl_data(rx_data_f2),
    .state(CRC_BYTES_st)
);

// CRC FSM
wire crc_feed;
wire crc_fetch_byte;
wire [TARGET_AW-1:0] crc_fetch_addr;

assign crc_enable_computation = misc_st[1];
assign crc_fetch_byte_sig = crc_fetch_byte;
wire crc_computation_in_progress = !crc_core_ready;

crcFSM #(
    .I_SIZE(I_SIZE),
    .TARGET_AW(TARGET_AW)
) fsm_crc (
    .clk(clk),
    .syn_rst(!soft_reset_b),
    .read_from_target(en_read),
    .I_st(CRC_BYTES_st[TARGET_AW-1:0]),
    .crc_feed(crc_feed),
    .start_compute_run_crc(crc_start_compute_run),
    .ready(crc_core_ready),
    .fetch_byte_target(crc_fetch_byte),
    .fetch_byte_addr(crc_fetch_addr)
);

// CRC core
reg feed_crc_delayed;
always@(posedge clk)
if(!soft_reset_b)
    feed_crc_delayed <= 0;
else
    feed_crc_delayed <= crc_feed;

wire crc_syn_rst = rx_done && (rx_data_cmd==runCMD);
wire [31:0] CRC_value;

wrap_crc32 #(.DWIDTH(8))
crc_core 
(
    .clk(clk),
    .feed(feed_crc_delayed),//feed_crc_delayed),
    .syn_rst(crc_syn_rst || !soft_reset_b),
    .data_in(T_RDATA_REG),
    .crc_o(CRC_value)
);

// Read CRC logic
wire read_crc_cmd_received = rx_done && (rx_data_cmd == read_crcCMD);
reg [15:0] crc_addr;

always@(posedge clk)
if(!soft_reset_b)
    crc_addr <= 0;
else if(read_crc_cmd_received)
    crc_addr <= rx_data_f1;

always@(*) begin
    if(crc_addr==16'd0)
        crc_byte_read = CRC_value[0 +: 8]; 
    else if(crc_addr==16'd1)
        crc_byte_read = CRC_value[8 +: 8];
    else if(crc_addr==16'd2)
        crc_byte_read = CRC_value[16 +: 8];
    else if(crc_addr==16'd3)
        crc_byte_read = CRC_value[24 +: 8];  
    else
        crc_byte_read = 8'hFF;
end

///////////
// output multiplexer
assign T_WDATA = (F_used == 0) ? ram_data_out : (
    (F_used == 1) ? prng_out : prng2_out   
);

// output address
assign T_ADDR = crc_computation_in_progress ? crc_fetch_addr : (rx_data_cmd == readCMD) ? rx_data_f1[0 +: TARGET_AW] : addr_reg_d;//(8'hFF-I_reg);

// output T_WRITE
//assign T_WRITE = en_I_reg_counter;
always@(posedge clk)begin
    T_WRITE <= en_I_reg_counter;
end


// output clock circuitry
generate
if (INV_OUT_CLK == 0)
    ODDR2 CB(.D0(1'b0),.D1(1'b1),.C0(~clk_target),.C1(clk_target),.Q(T_CLK),.CE(1'b1),.R(1'b0));
else
    ODDR2 CB(.D0(1'b0),.D1(1'b1),.C0(clk_target),.C1(~clk_target),.Q(T_CLK),.CE(1'b1),.R(1'b0));
endgenerate

wire [1:0] test = {soft_reset_b,poll_eng_auth};

assign C_GPIO = {3'b0,debug_clk_ctrl,1'b0,debug_clk_target,8'b0,test};
//assign C_GPIO = 0;
assign LED_STATUS = {misc_st[3:0],3'b0,misc_st[2]};

// POLL AUTHORITY BLOCK
always@(posedge clk)begin
    if (tx_done || (~soft_reset_b))
        poll_eng_auth <= 0;
    else if ((rx_done) && (rx_data_cmd == pollCMD))
        poll_eng_auth <= 1;
end

always@(posedge clk)begin
    if (~soft_reset_b)
        prev_state <= 8'hFF;
    else if ((rx_done) && (rx_data_cmd == pollCMD))
        prev_state <= {5'b11001,STATUS};
end

// Reset target
assign T_RSTB = (~soft_reset_b || (~reset_target_fsm && misc_st[0])) ? 1'b0 : 1'b1;

always @(posedge clk) begin
    if (~soft_reset_b) rnd_use_base <= 0;
    else if (sample_use_base_prng) rnd_use_base <= {rnd_use_base[0 +: 24], prng_out};
end

// Computation of offset (only 3 bits used)
always@(posedge clk) begin
    if (~soft_reset_b) ptr_state_index <= 0;
    else ptr_state_index <= rnd_use_base[23:0] % NSTATES_st[BITS_MODULO:0]; 
end

endmodule
