// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

// Low level Sakura API handler for UCL-Crypto. 
// Fetch all the input data from the controller in a shift register (only T_ADDR is not
// used). Assume thus that the incoming bytes are received "in-order".
module ucl_cg_sakura_api_handler
#(
    // Application parameters
    parameter BUFFER_IN_BYTES = 10,
    parameter BUFFER_OUT_BYTES = 10,
    parameter DELAY_PIPE_IN = 1
)
(
    // SAKURA-G API transaction signals
    input T_CLK, 
    input T_RSTB,
    input [7:0] T_WDATA,
    output [7:0] T_RDATA,
    input T_WRITE, 
    input T_VALID, 
    input T_START, // T_START to be sync with input to ensure that no race issue occurs
    output T_READY,
    // Top level (easy to use value)
    output [BUFFER_IN_BYTES*8-1:0] buffer_in,
    input [BUFFER_OUT_BYTES*8-1:0] buffer_out,
    input save_buffer_out,
    output start_execution,
    output api_clock,
    input done_execution,
    input rst_buffer_out,
    output rst
);

////// INPUT transactions
// In valid signal with pipeline.
// Pipeline use to stabilize input signal. Required for dense architecture
// in order to avoid congestion at the area close to the IOs of the FPGA.
localparam BITS_IN_PIPE = 12;
wire [BITS_IN_PIPE-1:0] pin_stable_in = {T_START, T_WRITE, T_VALID, T_RSTB, T_WDATA}; 
wire [BITS_IN_PIPE-1:0] pout_stable_in;

pipe_reg #(
    .W(BITS_IN_PIPE),
    .DEPTH(DELAY_PIPE_IN)
) pipe_stable_in (
    .clk(T_CLK),
    .in(pin_stable_in),
    .out(pout_stable_in)
);

// Fetch stable data from pipeline
wire [7:0] sT_WDATA = pout_stable_in[0 +: 8];
wire sT_RSTB = pout_stable_in[8];
wire sT_VALID = pout_stable_in[9];
wire sT_WRITE = pout_stable_in[10];
wire sT_START = pout_stable_in[11];

// Control signal based on stable input
wire nrst = sT_RSTB;
wire fetch_in = sT_VALID & sT_WRITE;
assign start_execution = sT_START;
assign rst = ~nrst;

// Buffer for the input
reg [BUFFER_IN_BYTES*8-1:0] reg_buffer_in;
always@(posedge T_CLK)
if(~nrst) begin
    reg_buffer_in <= 0;
end else if(fetch_in) begin
    reg_buffer_in <= {sT_WDATA, reg_buffer_in[BUFFER_IN_BYTES*8-1:8]};
end else begin
    reg_buffer_in <= reg_buffer_in;
end
assign buffer_in = reg_buffer_in;

//////// OUTPUT transactions
// Should directly interact with the control API (delay should be ensured and cannot be delayed)
wire fetch_out = T_VALID & ~T_WRITE;

reg [BUFFER_OUT_BYTES*8-1:0] reg_buffer_out; 
always@(posedge T_CLK) 
if(rst_buffer_out) begin
    reg_buffer_out <= 0;
end else if(save_buffer_out) begin
    reg_buffer_out <= buffer_out; 
end else if(fetch_out) begin
    reg_buffer_out <= {8'b0, reg_buffer_out[BUFFER_OUT_BYTES*8-1:8]};
end else begin
    reg_buffer_out <= reg_buffer_out;
end

reg [7:0] reg_stable_out;
always@(posedge T_CLK) begin
    reg_stable_out <= reg_buffer_out[7:0];
end
assign T_RDATA = reg_stable_out;

// Output clock
assign api_clock = T_CLK;

/////// Register to set the T_READY value ath the end of the execution
reg reg_T_READY;
always@(posedge T_CLK) begin
    reg_T_READY <= done_execution;
end
assign T_READY = reg_T_READY;


endmodule
