// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module ucl_cg_sakura_dut_sr
#(
    parameter BUFFER_IN_BYTES = 10,
    parameter BUFFER_OUT_BYTES = 10
    // Add your custom parameter here
    // DO NOT MODIFY THE PORTS
)
(
    input api_clock,
    input rst,
    input [BUFFER_IN_BYTES*8-1:0] buffer_in,
    output [BUFFER_OUT_BYTES*8-1:0] buffer_out,
    output save_buffer_out,
    output rst_buffer_out,
    input start_execution,
    output done_execution,
    output [3:0] LED_STATUS,
    output [15:0] T_GPIO
);

//`define HPC4
`define HPC3

// Simple FSM:
// 1) [Handled by api_handler] Fetch the data required for an execution
// 2) Wait for start_execution
// 3) Wait for DELAY_CYCLES cycles
// 4) Save the value received as the output buffer
// 5) Raise the ending of the execution
// 6) [Handle by api handler]
localparam COMPUTE_CYCLES = 2;
localparam DELAY_CYCLES = 5; 
localparam d = 2;
localparam and_hpc3_nrnd = 2*d*(d-1)/2;
localparam and_hpc4_nrnd = 5*d*(d-1)/2;
localparam B = 8;
localparam I = 1;

`ifdef HPC4
    localparam and_nrnd = and_hpc4_nrnd;
`else
    localparam and_nrnd = and_hpc3_nrnd;
`endif

// FSM state
localparam IDLE=0,
    WAIT_PRE_TRIGGER=1,
    WAIT_COMPUTE=2,
    SAVE_OUTPUT=3,
    WAIT_POST_TRIGGER=4,
    ASSERT_DONE=5,
    UNREACHABLE=6;

reg [3:0] state, nextstate;
always@(posedge api_clock)
if(rst) begin
    state <= IDLE;
end else begin
    state <= nextstate;
end

// delay counter
reg [5:0] fsm_cnt;
reg rst_cnt, inc_cnt;
reg [5:0] data_cnt;
reg rst_data_cnt, inc_data_cnt;

always@(posedge api_clock)
if(rst_cnt) begin
    fsm_cnt <= 0;
end else if(inc_cnt) begin
    fsm_cnt <= fsm_cnt + 1;
end else begin
    fsm_cnt <= fsm_cnt;
end

always@(posedge api_clock)
if(rst_data_cnt) begin
    data_cnt <= 0;
end else if(inc_data_cnt) begin
    data_cnt <= data_cnt + 1;
end else begin
    data_cnt <= data_cnt;
end

wire delay_done = (fsm_cnt == DELAY_CYCLES-1);

// Control reg
reg flag_trigger, reg_flag_trigger;
reg reg_save_buffer_out;
reg reg_done_execution;
reg reg_rst_buffer_out;
reg en_rnd;

reg [B*2-1:0] tmp_buffer_out;

assign T_GPIO = {10'b0, reg_flag_trigger, 1'b0, reg_flag_trigger, 3'b0};

// Nextstate logic
always@(*) begin
    // API control 
    reg_save_buffer_out = 0;
    reg_done_execution = 0;
    reg_rst_buffer_out = 0;

    // Custom control
    nextstate = state;
    rst_cnt = 0;
    rst_data_cnt = 0;
    inc_cnt = 0;
    inc_data_cnt = 0;
    flag_trigger = 1;
    en = 0;
    en_rnd = 0;
    case(state)
        IDLE: begin
            flag_trigger = 0;
            rst_cnt = 1;
            rst_data_cnt = 1;
            if(start_execution) begin
                nextstate = WAIT_PRE_TRIGGER;
            end
        end
        WAIT_PRE_TRIGGER: begin
            inc_cnt = 1;
            rst_data_cnt = 1;
            if(fsm_cnt-1==DELAY_CYCLES) begin
                inc_data_cnt = 1;
                rst_data_cnt = 0;
            end
            if(delay_done) begin
                nextstate = WAIT_COMPUTE;
                rst_cnt = 1;
                rst_data_cnt = 0;
                inc_data_cnt = 1;
                en = 1;
                en_rnd = 1;
            end
        end
        WAIT_COMPUTE: begin
            inc_cnt = 1;
            inc_data_cnt = 1;
            en_rnd = 1;
            if(en_FFF) begin
                nextstate = SAVE_OUTPUT;
                rst_cnt = 1;
            end
        end
        SAVE_OUTPUT: begin
            nextstate = WAIT_POST_TRIGGER;
            reg_save_buffer_out = 1;
            rst_cnt = 1;
        end
        WAIT_POST_TRIGGER: begin 
            nextstate = WAIT_POST_TRIGGER;
            inc_cnt = 1;
            flag_trigger = 0;
            if(delay_done) begin
                nextstate=ASSERT_DONE;
            end
        end
        ASSERT_DONE: begin
            nextstate = IDLE;
            reg_done_execution = 1;
            flag_trigger = 0;
        end
        default: nextstate=UNREACHABLE;
    endcase
end

// Bus attribution

wire [d-1:0] gin_x, gin_y;
reg [d-1:0] reg_x, reg_y;

reg [and_nrnd-1:0] reg_rnd;
wire [and_nrnd-1:0] gin_rnd;

wire [I*d-1:0] z_all;
wire [d-1:0] z;

wire [and_nrnd-1:0] buffer_rnd;
(* KEEP = "TRUE", S = "TRUE", KEEP_HIERARCHY = "TRUE", DONT_TOUCH = "TRUE" *)
wire [d-1:0] to_keep_lolilol [I-1:0];

wire dummy_out;

reg en, en_F, en_FF, en_FFF;
always@(posedge api_clock) begin 
    reg_flag_trigger <= flag_trigger;
    en_F <= en;
    en_FF <= en_F;
    en_FFF <= en_FF;
    if(rst|en_FFF) begin
        reg_x <= 0;
        reg_y <= 0;
    end
    else if (en) begin
        reg_x <= {buffer_in[6*B], buffer_in[9*B]};
        reg_y <= {buffer_in[7*B], buffer_in[10*B]};
    end
    else begin
        reg_x <= reg_x;
        reg_y <= reg_y;
    end
end

always@(posedge api_clock) begin 
    if(rst|en_FFF) begin
        reg_rnd <= 0;
    end
    else if (en_rnd) begin
        reg_rnd <= buffer_rnd;
    end
    else begin
        reg_rnd <= reg_rnd;
    end
end

assign gin_y = reg_y;
assign gin_x = en_FF ? z : reg_x;
assign gin_rnd = reg_rnd;
// Mux the randomness
assign buffer_rnd = buffer_in[data_cnt*2*B+:and_nrnd];


genvar i;
generate
    for (i=0; i < I; i = i+1) begin : gadgets
        `ifdef HPC4
            //assign z_all[d*(i+1)-1:d*i] = x & y;
	    (* KEEP_HIERARCHY = "TRUE" *)
	    MSKand_HPC4 #(.d(d)) gadgeti (.ina(gin_x), .inb(gin_y), .rnd(gin_rnd), .clk(api_clock),.out(z_all[d*(i+1)-1:d*i]));
        `else
		     	(* KEEP_HIERARCHY = "TRUE" *)
            MSKand_HPC3 #(.d(d)) gadgeti (.ina(gin_x), .inb(gin_y), .rnd(gin_rnd), .clk(api_clock),.out(z_all[d*(i+1)-1:d*i]));
				//and_HPC3 #(.security_order(d)) gadgeti (.a(x), .b(y), .r(rnd), .clk(api_clk), .c(z_all[d*(i+1)-1:d*i]));
        `endif
        assign to_keep_lolilol[i] = z_all[d*(i+1)-1:d*i];
    end
endgenerate

assign z = z_all[d-1:0];

assign buffer_out = buffer_in;
assign save_buffer_out = reg_save_buffer_out;
assign done_execution = reg_done_execution;
assign rst_buffer_out = reg_rst_buffer_out;

// Set the LEDs to the constant 0
assign LED_STATUS = 4'b0; 

endmodule
