// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module levelled_datapath #(parameter d = 2) (
    // CLK
    clk, 
    // FLAGS
    flag_dp_init_dpa,
    flag_dp_sel_tgf_in,
    flag_dp_sel_tgf_out,
    flag_dp_sel_kdf,
    flag_dp_decrypt,
    flag_dp_stall,
    flag_dp_unstall,
    flag_dp_unmask,
    flag_dp_plout,
    flag_dp_cst,
    flag_dp_transition,
    //      DPA core
    flag_dpa_reset,
    flag_dpa_start,
    flag_dpa_last,
    flag_dpa_done,
    flag_dp_perm,
    flag_dp_mask,
    flag_spa_reset,
    flag_spa_start,
    flag_spa_last,
    flag_spa_done,
    flag_spa_perm,
    // PAYLOADS
    pl_dp_in,
    pl_dp_nonce,
    pl_dp_k,
    pl_dp_rnd,
    pl_dp_tag,
    pl_dp_out
    );

    // PARAMETER DEFINITION
    parameter and_pini_nrnd = 5*d*(d-1)/2;
    parameter sbox_pini_nrnd = 5*and_pini_nrnd;
    parameter slayer_pini_nrnd = 64*sbox_pini_nrnd;

    // CLK
    input clk;
    // FLAGS
    input flag_dp_init_dpa;
    input flag_dp_sel_tgf_in;
    input flag_dp_sel_tgf_out;
    input flag_dp_sel_kdf;
    input flag_dp_decrypt;
    input flag_dp_stall;
    input flag_dp_unstall;
    input flag_dp_unmask;
    input flag_dp_plout;
    input flag_dp_cst;
    output flag_dp_transition;
    // DPA
    input flag_dpa_reset;
    input flag_dpa_start;
    input flag_dp_perm;
    input flag_dp_mask;
    input flag_dpa_last;
    output flag_dpa_done;
    // SPA
    input flag_spa_reset;
    input flag_spa_start;
    input flag_spa_last;
    output flag_spa_done;
    input flag_spa_perm;
    // PAYLOAD
    input [0:64*d-1] pl_dp_in;
    input [0:128-1] pl_dp_nonce;
    input [0:128*d-1] pl_dp_k;
    input [0:slayer_pini_nrnd-1] pl_dp_rnd;
    output [0:128-1] pl_dp_tag;
    output [0:64*d-1] pl_dp_out;



    // Mode REG & WIRE
    reg flag_dp_unmask_F;
    wire [0:64-1] u_pl_dp_in, u_pl_dp_out;
    wire [0:d-1] m_flag_dp_cst;
    wire [0:64*d-1] m_iv, pl_dp_out_dpa, m_pl_dp_out_spa;
    wire [0:128*d-1] m_nonce, m_tag;
    wire [0:320-1] state_umsk, state_umsk_stall, state_out_spa_cst, state_umsk_mux;
    wire [0:320*d-1] state_init_dpa, m_state_out_spa, state_mux1_dpa, state_out_dpa, state_out_dpa_kdf_bulk, state_out_dpa_kdf_cst, state_out_dpa_kdf, state_out_computed, state_in_dpa, state_out_mode_dpa, state_mux_dpa, state_mux_dpa_tgf;
    // DPA core REG & WIRE
    reg flag_dpa_perm, flag_dpa_mask;
    wire [0:320*d-1] pl_dpa_in, pl_dpa_out;
    wire flag_dpa_transition, flag_dpa_done;
    // SPA core REG & WIRE
    wire [0:320-1] pl_spa_in, pl_spa_out, state_out_spa;
    wire flag_spa_transition;
    // DPA path
    //      Initial values
    MSKcst #(.d(d), .count(64)) msk_iv (64'h80400c0600000000, m_iv);
    MSKcst #(.d(d), .count(128)) msk_nonce (pl_dp_nonce, m_nonce);
    assign state_init_dpa = {m_iv, pl_dp_k, m_nonce};

    MSKmux #(.d(d), .count(320)) mux_init2 (flag_dp_mask, state_out_dpa_kdf_bulk, m_state_out_spa, state_mux1_dpa);
    MSKmux #(.d(d), .count(320)) mux_init (flag_dp_init_dpa, state_init_dpa, state_mux1_dpa, state_mux_dpa);

    //      TGF addition
    MSK_tgf_in #(.d(d)) TGF_in (clk, flag_dp_sel_tgf_in, pl_dp_k, state_mux_dpa, state_mux_dpa_tgf);

    //      Crypto protected core
    DPA_Ascon_FSM #(.d(d)) ascon_dpa(
        // CLK
        .clk(clk),                    // clock
        // FLAGS
        .flag_dpa_start,         // start core from IDLE
        .flag_dpa_perm(flag_dp_perm),          // indicate perm to do perm_a / perm_b
        .flag_dpa_mask,          // indicate if dpa countermeasure should be on/off
        .flag_dpa_reset,         // reset core
        .flag_dpa_last,          // indicate that last operation has been flag_dpa_done
        .flag_dpa_transition,        // indicate that cycle before output has been flag_dpa_done
        .flag_dpa_done,          // indicate that operation is flag_dpa_done
        // PAYLOADS
        .pl_dpa_in(state_mux_dpa_tgf),              // input state
        .pl_dpa_rnd(pl_dp_rnd),             // input randomness
        .pl_dpa_out(state_out_dpa)              // output state
    );
    MSKxor #(.d(d), .count(128)) xor_tag (pl_dp_k, state_out_dpa[192*d:320*d-1], m_tag);
    
    //      KDF addition
    MSK_kdf #(.d(d)) KDF (clk, flag_dp_sel_kdf, pl_dp_k, state_out_dpa, state_out_dpa_kdf);
    //      Cst addition
    MSKcst #(.d(d), .count(1)) domain_separation (flag_dp_cst, m_flag_dp_cst);
    MSKxor #(.d(d), .count(1)) xor_domain_cst (state_out_dpa_kdf[319*d:320*d-1], m_flag_dp_cst, state_out_dpa_kdf_cst[319*d:320*d-1]);
    assign state_out_dpa_kdf_cst[0:319*d-1] = state_out_dpa_kdf[0:319*d-1];
    //      Stall and absorption

    MSK_stall_abs #(.d(d)) DPA_Bulk (clk, flag_dp_stall, flag_dp_unstall, flag_dp_decrypt, pl_dp_in, state_out_dpa_kdf_cst, state_out_dpa_kdf_bulk, pl_dp_out_dpa);  
    //      Unmask
    genvar i;
    generate
        for(i=0; i<320; i=i+1) begin
            assign state_umsk[i] = ^(state_out_dpa_kdf[d*i:d*(i+1)-1] & flag_dp_unmask);
        end
        for(i=0; i<128; i=i+1) begin
            assign pl_dp_tag[i] = ^(m_tag[d*i:d*(i+1)-1] & flag_dp_unmask);
        end
        for(i=0; i<64; i=i+1) begin
            assign u_pl_dp_in[i] = ^(pl_dp_in[d*i:d*(i+1)-1] & flag_dp_unmask);
        end
    endgenerate
    always @(posedge clk ) begin
        flag_dp_unmask_F <= flag_dp_unmask;
    end
    //      Stall and absorption  
    assign state_out_spa_cst[0:320-2] = state_out_spa[0:320-2];
    assign state_out_spa_cst[320-1] = state_out_spa[320-1]^flag_dp_cst;
    assign state_umsk_mux = (flag_dp_unmask_F) ? state_umsk : state_out_spa_cst;
    UMSK_stall_abs Bulk(clk, flag_dp_stall, flag_dp_unstall, flag_dp_decrypt, u_pl_dp_in, state_umsk_mux, state_umsk_stall, u_pl_dp_out);
    MSKcst #(.d(d), .count(64)) dummy_payload (u_pl_dp_out, m_pl_dp_out_spa);
    //      Crypto unprotected core
    MSKcst #(.d(d), .count(320)) cst_spaout (state_umsk_stall, m_state_out_spa);

    SPA_Ascon_FSM ascon_spa (
        .flag_spa_start,
        .clk,
        .flag_spa_reset,
        .flag_spa_last,
        .pl_spa_in(state_umsk_stall),
        .flag_spa_done,
        .flag_spa_transition,
        .pl_spa_out(state_out_spa)
    );
    assign flag_dp_transition = flag_spa_transition | flag_dpa_transition;
    //      Domain separation
    MSKmux #(.d(d), .count(64)) mux_payload (flag_dp_mask, pl_dp_out_dpa, m_pl_dp_out_spa, pl_dp_out);
endmodule