// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module levelled_FSM #(parameter d = 2, parameter set_prng = 0) (
    // CLK
    clk,            // Clock input
    // FLAG SIGNALS
    flag_core_reset,          // flag_core_reset signal
    flag_core_decrypt,        // Decrypt flag
    flag_core_last_pt,
    flag_core_ad,        // Flag to indicate that the spa input is AD
    flag_core_confidentiality,
    flag_core_integrity,
    flag_core_done,
    // AXI
    flag_core_in_spa_valid,   // Master is providing valid data for SPA
    flag_core_in_spa_ready, // Slave is ready to process data for SPA
    flag_core_in_dpa_valid,   // Master is providing valid data for DPA
    flag_core_in_dpa_ready,   // Slave is ready to process data for DPA
    flag_core_out_valid,      // Slave is providing valid data
    flag_core_out_ready,      // Master is ready to process data
    // PAYLOAD
    pl_core_spa,             // 128 input plaintext bus
    pl_core_k,           // 128*d input key bus
    pl_core_nonce,       // 128 input nonce bus
    pl_core_seed,
    pl_core_comp_spa,
    pl_core_tag
);

    // CLK
    input clk;
    // FLAG
    input flag_core_reset;
    input flag_core_decrypt;        // Decrypt flag
    input flag_core_last_pt;
    input flag_core_ad;        // Flag to indicate that the spa input is AD
    input flag_core_confidentiality;
    input flag_core_integrity;
    output reg flag_core_done;
    // AXI
    input flag_core_in_spa_valid;   // Master is providing valid data for SPA
    output reg flag_core_in_spa_ready; // Slave is ready to process data for SPA
    input flag_core_in_dpa_valid;   // Master is providing valid data for DPA
    output reg flag_core_in_dpa_ready;   // Slave is ready to process data for DPA
    output reg flag_core_out_valid;      // Slave is providing valid data
    input flag_core_out_ready;      // Master is ready to process data
    // PAYLOAD
    input [0:64-1] pl_core_spa;             // 128 input plaintext bus
    input [0:128*d-1] pl_core_k;           // 128*d input key bus
    input [0:128-1] pl_core_nonce;       // 128 input nonce bus
    input [0:80-1] pl_core_seed;
    output [0:64-1] pl_core_comp_spa;
    output [0:128-1] pl_core_tag;

    // STATE DEFINITION
    `define IDLE        0
    `define PRNG        1
    `define KDF         2
    `define TGF         3
    `define AD          4
    `define PT          5
    `define STALL       7

    // PARAMETER DEFINITION
    parameter and_pini_nrnd = 5*d*(d-1)/2;
    parameter sbox_pini_nrnd = 5*and_pini_nrnd;
    parameter slayer_pini_nrnd = 64*sbox_pini_nrnd;


    // Control REG & WIRE
    reg [0:2] state, next_state;
    reg flag_core_last_pt_F, flag_dp_cst, flag_core_last_pt_FF, flag_init_dpa;
    wire flag_core_last_pt_rotten, flag_dpa_last, flag_stall_ad, flag_stall_pt;
    // Datapath REG & WIRE
    reg flag_dp_init_dpa, flag_dp_sel_kdf, flag_dp_sel_tgf_in, flag_dp_sel_tgf_out;
    reg flag_dpa_start, flag_dp_unmask, flag_dp_perm, flag_dpa_reset, flag_dp_stall, flag_dp_unstall, flag_dp_plout, flag_dp_mask;
    wire flag_dp_decrypt;
    wire flag_dp_transition, flag_dpa_done;
    reg flag_spa_start, flag_spa_reset, flag_spa_last, flag_spa_perm;
    wire flag_spa_done;
    // PRNG REG & WIRE
    wire [0:80-1] prng_seed;
    wire prng_busy, prng_out_valid;
    reg prng_reset, prng_start_reseed, prng_out_ready;
    wire [0:64*5*and_pini_nrnd-1] pl_prng_rnd;
    wire [0:64*5*and_pini_nrnd-1] pl_dp_rnd;

    // PRNG

    assign prng_seed = pl_core_seed;
    
    prng_top #(.RND(64*25*d*(d-1)/2), .NINIT(4*288), .MAX_UNROLL(128)) trivium1 (
        prng_reset,
        clk,            // Clock
        prng_start_reseed,   // Acive high. Start the reseeding procedure.
        // Active high, reseed in progress,
        prng_busy,
        // The seed, loaded as initial state.
        prng_seed,
        // SVRS pseudo-random output stream.
        // Output will be always valid, except:
        // - when a reseeding is in progress
        // - when the PRNG has not been reseeded yet after a flag_core_reset.
        prng_out_ready,
        prng_out_valid,
        pl_prng_rnd
    );

    if (set_prng == 1) begin
            assign pl_dp_rnd = pl_prng_rnd;
        end
    else begin
        assign pl_dp_rnd = {(64*5*and_pini_nrnd){1'b0}};
    end

    // Datapath
    assign flag_dp_decrypt = flag_core_decrypt;
    levelled_datapath #(.d(d)) datapath (
        // CLK
        .clk(clk), 
        // FLAGS
        //      Mode
        .flag_dp_init_dpa,
        .flag_dp_sel_tgf_in,
        .flag_dp_sel_tgf_out,
        .flag_dp_sel_kdf,
        .flag_dp_decrypt,
        .flag_dp_stall,
        .flag_dp_unstall,
        .flag_dp_unmask,
        .flag_dp_plout,
        .flag_dp_cst,
        .flag_dp_transition,
        //      DPA core
        .flag_dpa_reset,
        .flag_dpa_start,
        .flag_dp_perm,
        .flag_dp_mask,
        .flag_dpa_last,
        .flag_dpa_done,
        //      SPA core
        .flag_spa_reset,
        .flag_spa_start,
        .flag_spa_last,
        .flag_spa_done,
        .flag_spa_perm,
        // PAYLOADS
        .pl_dp_in(pl_core_spa),
        .pl_dp_nonce(pl_core_nonce),
        .pl_dp_k(pl_core_k),
        .pl_dp_rnd(pl_prng_rnd),
        .pl_dp_tag(pl_core_tag),
        .pl_dp_out(pl_core_comp_spa)
    );

always @(*) begin
    next_state = state;
    // PRNG
    prng_reset = 1;
    prng_start_reseed = 0;
    prng_out_ready = 0;
    // DPA
    flag_dpa_reset = 1;
    flag_dpa_start = 0;
    flag_dp_perm = 0;
    // SPA
    flag_spa_reset = 1;
    flag_spa_start = 0;
    flag_spa_perm = 0;
    // Mode
    flag_dp_sel_kdf = 0;
    flag_dp_sel_tgf_in = 0;
    flag_dp_sel_tgf_out = 0;
    flag_dp_unstall = 0;
    flag_dp_stall = 0;
    flag_core_done = 0;
    flag_dp_cst = 0;
    flag_dp_unmask = 0;
    flag_dp_init_dpa = 0;
    flag_dp_plout = 0;

    // AXI
    flag_core_in_dpa_ready = 0;
    flag_core_in_spa_ready = 0;
    flag_core_out_valid = 0;

    case (state) 
        `IDLE : begin
            if (flag_core_in_dpa_valid) begin
                next_state = `PRNG;
                prng_reset = 0;
                prng_start_reseed = 1;
            end
        end
        `PRNG : begin
            prng_reset = 0;
            // When PRNG is outputing randomness, use it directly
            if(prng_out_valid) begin
                next_state = `KDF;
                flag_dpa_start = 1;
                flag_dpa_reset = 0;
                prng_out_ready = 1;     
                flag_dp_init_dpa = 1;
                flag_dp_perm = 1;
            end
        end
        `KDF : begin
            prng_reset = 0;
            flag_dpa_reset = 0;
            prng_out_ready = 1;
            if (flag_dp_transition) begin
                flag_dp_sel_kdf = 1;
                flag_dp_unmask = 1;
                if (flag_core_last_pt) begin
                    flag_dp_sel_tgf_in = 1;
                end
            end
            if (flag_dpa_done) begin
                if(flag_core_ad) begin
                    next_state = `AD;
                    flag_spa_reset = 0;
                    flag_spa_start = 1;
                end
                else if (flag_core_last_pt) begin
                    next_state = `TGF;
                    flag_dp_cst = 1;
                    flag_dp_perm = 1;
                end
                else begin
                    next_state = `PT;
                    flag_dp_cst = 1;
                    flag_spa_reset = 0;
                    flag_spa_start = 1;
                end
            end
        end
        `AD : begin
            flag_spa_reset = 0;
            prng_reset = 0;
            if (flag_dp_transition) begin
                if (~flag_core_ad)begin 
                    flag_dp_plout = 1;
                end
                if (flag_core_last_pt) begin
                    flag_dp_sel_tgf_out = 1;
                    flag_dp_unmask = 1;
                end
            end
            if(flag_dpa_done) begin
                if (flag_core_ad) begin
                end
                else if (flag_core_last_pt & ~flag_core_last_pt_rotten) begin
                    next_state = `TGF;
                    flag_dp_cst = 1;
                    flag_dpa_reset = 0;
                    flag_dpa_start = 1;
                    flag_dp_perm = 1;
                end
                else if (flag_core_last_pt) begin
                    next_state = `STALL;
                    flag_dp_sel_tgf_out = 1;
                    flag_dp_stall = 1;
                end
                else begin
                    next_state = `PT;
                    flag_dp_cst = 1; 
                    flag_spa_reset = 0;
                    flag_spa_start = 1;
                    flag_dp_plout = 1;
                end
            end
        end
        `PT : begin
            prng_reset = 0;
            flag_spa_reset = 0;
            if (flag_dp_transition) begin
                flag_dp_plout = 1;
                if (flag_core_last_pt & ~flag_core_last_pt_rotten) begin
                    flag_dp_sel_tgf_in = 1;
                    flag_dp_unmask = 1;
                end
            end
            if (flag_spa_done) begin
                if (flag_stall_pt) begin
                    next_state = `STALL;
                    flag_dp_stall = 1;
                end
                else begin
                    if(flag_core_last_pt & ~flag_core_last_pt_rotten) begin
                        next_state = `TGF;
                        flag_dpa_reset = 0;
                        flag_dpa_start = 1;
                        flag_dp_perm = 1;
                    end
                    else if (flag_core_last_pt) begin
                        next_state = `STALL;
                        flag_dp_stall = 1;
                    end
                end
            end
        end
        `TGF : begin
            prng_out_ready = 1;
            prng_reset = 0;
            flag_dpa_reset = 0;
            if (flag_dp_transition) begin
                flag_dp_sel_tgf_out = 1;
            end
            if(flag_dpa_done) begin
                next_state = `IDLE;
                flag_core_done = 1;
            end
        end
        `STALL : begin
            if (flag_core_last_pt_FF) begin
                next_state = `TGF;
                flag_dp_cst = 1;
                flag_dp_plout = 1;
                flag_dpa_reset = 0;
                flag_dpa_start = 1;
                flag_dp_unstall = 1;
                prng_out_ready = 1;
            end
        end
    endcase
end

assign flag_dpa_last = (state==`KDF) | (state==`TGF) & flag_core_last_pt;
assign flag_core_last_pt_rotten = ~flag_core_last_pt_F & flag_core_last_pt;
assign flag_stall_ad = ~flag_core_in_spa_valid;
assign flag_stall_pt = ~flag_core_in_spa_valid | ~flag_core_out_ready;
always @(posedge clk ) begin
    flag_core_last_pt_F <= flag_core_last_pt;
    flag_core_last_pt_FF <= flag_core_last_pt_F;
end

always @(posedge clk) begin
    if (flag_core_reset) begin
        state <= `IDLE;
    end
    else begin
        state <= next_state;
    end
end 
endmodule