// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
(*fv_prop = "PINI", fv_strat = "composite", fv_order=d*)
module DPA_Ascon_FSM #(parameter d = 2, set_prng = 0) (
    clk,
    flag_dpa_start,
    flag_dpa_perm,
    flag_dpa_mask,
    flag_dpa_reset,
    flag_dpa_last,
    flag_dpa_transition,        // indicate that cycle before output has been flag_dpa_done
    flag_dpa_done,
    pl_dpa_in,
    pl_dpa_rnd,
    pl_dpa_out
);
    //`include "MSKand_HPC2.vh"
    localparam and_pini_mul_nrnd = d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;
    // INPUT
    (* fv_type="clock" *) input clk;
    (* fv_type="control" *) input flag_dpa_start, flag_dpa_perm, flag_dpa_mask, flag_dpa_reset, flag_dpa_last;
    output reg flag_dpa_transition;
    output reg flag_dpa_done;
    (* fv_type="sharing", fv_latency=0, fv_count=320 *) input [320*d-1 : 0] pl_dpa_in;
    (* syn_keep = "true", keep = "true", fv_type = "random", fv_latency=0, fv_count = 1, fv_rnd_count_0 = 80, fv_rnd_lat_0 = 0 *) input [16*5*and_pini_nrnd-1:0] pl_dpa_rnd;

    // OUTPUT
    (* fv_type="sharing", fv_latency=240, fv_count=320 *) output [320*d-1:0] pl_dpa_out;
    // REGS
    reg start_round, reset_round_cnt;
    reg sel1, sel2, sel_cst;
    reg [1:0] state, next_state;
    reg [3:0] cycle_cnt;
    reg [5:0] round_cnt;

    // WIRE 
    wire [320*d-1:0] pl_dpa_out;

    // STATE DEFINITION
    `define IDLE 0
    `define flag_dpa_perm 1
    `define PERM_B 2

    // INSTANTIATION
    DPA_Ascon_datapath #(d) datapath (clk, sel1, sel2, sel_cst, flag_dpa_done, pl_dpa_rnd, pl_dpa_in, pl_dpa_out);

always @(posedge clk) begin
    if (flag_dpa_reset) begin
        state <=  `IDLE;
    end
    else begin
        state <= next_state;
    end
end 

always @(*) begin
    next_state = state;
    flag_dpa_done = 0;
    sel1 = 0;
    sel2 = 0;
    start_round = 0;
    sel_cst = 0;
    reset_round_cnt = 0;
    flag_dpa_transition = 0;
    case (state)
        `IDLE : begin
            if (flag_dpa_start) begin
                if (flag_dpa_perm) begin
                    next_state = `flag_dpa_perm;
                end
                else begin
                    next_state = `PERM_B;
                end
                reset_round_cnt = 1;
                sel1 = 1;
                sel2 = 1;
            end
        end
        `flag_dpa_perm : begin
            if (cycle_cnt == 0) begin
                sel_cst = 1;
            end
            if (cycle_cnt == 5) begin
               sel1 = 0; 
               sel2 = 1;
               start_round = 1;
            end
            if (round_cnt == 11 & cycle_cnt == 4) begin
                flag_dpa_transition = 1;
            end
            if (round_cnt == 11 & cycle_cnt == 5 & flag_dpa_last) begin
                next_state = `IDLE;
                flag_dpa_done = 1;
                sel1 = 1;
                sel2 = 1;
                reset_round_cnt = 1;
            end
            else if (round_cnt == 11 & cycle_cnt == 5 & ~flag_dpa_perm) begin
                next_state = `PERM_B;
                flag_dpa_done = 1;
                sel1 = 1;
                sel2 = 1;
                reset_round_cnt = 1;
            end
            else if (round_cnt == 11 & cycle_cnt == 5) begin
                flag_dpa_done = 1;
                sel1 = 1;
                sel2 = 1;
                reset_round_cnt = 1;
            end
        end
        `PERM_B : begin
            if (cycle_cnt == 0) begin
                sel_cst = 1;
            end
            if (cycle_cnt == 5) begin
                sel1 = 0; 
                sel2 = 1;
                start_round = 1;
            end
            if (round_cnt == 5 & cycle_cnt == 4) begin
                flag_dpa_transition = 1;
            end
            if (round_cnt == 5 & cycle_cnt == 5 & flag_dpa_last) begin
                next_state = `IDLE;
                flag_dpa_done = 1;
                sel1 = 1;
                sel2 = 1;
                reset_round_cnt = 1;
            end
            else if (round_cnt == 5 & cycle_cnt == 5 & flag_dpa_perm) begin
                next_state = `flag_dpa_perm;
                flag_dpa_done = 1;
                sel1 = 1;
                sel2 = 1;
                reset_round_cnt = 1;
            end
            else if (round_cnt == 5 & cycle_cnt == 5) begin
                flag_dpa_done = 1;
                sel1 = 1;
                sel2 = 1;
                reset_round_cnt = 1;
            end
        end
    endcase
end

always @(posedge clk) begin
    if((state == `PERM_B | state == `flag_dpa_perm) & ~start_round) begin
        cycle_cnt <= cycle_cnt + 1;
    end
    else begin
        cycle_cnt <= 0;
    end
end

// ROUND COUNTER EVOLUTION
always @(posedge clk) begin
    if(reset_round_cnt) begin
        round_cnt <= 0;
    end else if (start_round) begin
        round_cnt <= round_cnt + 1;
    end
end

endmodule