// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module DPA_Ascon_FSM #(parameter d = 2, set_prng = 0) (
    // CLK
    clk,                    // clock
    // FLAGS
    flag_dpa_start,         // start core from IDLE
    flag_dpa_perm,          // indicate perm to do perm_a / perm_b
    flag_dpa_mask,          // indicate if dpa countermeasure should be on/off
    flag_dpa_reset,         // reset core
    flag_dpa_last,          // indicate that last operation has been flag_dpa_done
    flag_dpa_transition,        // indicate that cycle before output has been flag_dpa_done
    flag_dpa_done,          // indicate that operation is flag_dpa_done
    // PAYLOADS
    pl_dpa_in,              // input state
    pl_dpa_rnd,             // input randomness
    pl_dpa_out              // output state
);
    // CLK
    input clk;
    // FLAGS
    input flag_dpa_start;
    input flag_dpa_perm;
    input flag_dpa_mask;
    input flag_dpa_reset;
    input flag_dpa_last;
    output reg flag_dpa_transition;
    output reg flag_dpa_done;
    // PAYLOADS
    input [0:320*d-1] pl_dpa_in;
    input [0:64*25*d*(d-1)/2-1] pl_dpa_rnd;
    // OUTPUT
    output [0:320*d-1] pl_dpa_out;

    // Control REG & WIRE
    reg reset_round_cnt;
    reg [0:1] state, next_state;
    reg [0:3] round_cnt;

    // Datapath REG & WIRE
    reg flag_dp_perm, flag_dp_start;
    wire [0:320*d-1] pl_dp_in, pl_dp_out;
    wire [0:64*25*d*(d-1)/2-1] pl_dp_rnd;
    // STATE DEFINITION
    `define IDLE 0
    `define PERM_A 1
    `define PERM_B 2

    assign pl_dp_in = pl_dpa_in;
    assign pl_dp_rnd = pl_dpa_rnd;

    // INSTANTIATION
    DPA_Ascon_datapath #(.d(d)) datapath (
        clk, 
        flag_dp_start, 
        flag_dp_perm, 
        pl_dp_in, 
        pl_dp_rnd, 
        pl_dp_out
    );

    assign pl_dpa_out = pl_dp_out;
    
always @(*) begin
    next_state = state;
    // Control
    reset_round_cnt = 0;

    // Datapath flags
    flag_dp_start = 0;
    flag_dpa_done = 0;
    flag_dp_perm = 0;
    flag_dpa_transition = 0;

    case (state)
        `IDLE : begin
            reset_round_cnt = 1;
            if (flag_dpa_start) begin
                if(flag_dpa_perm) begin
                    next_state = `PERM_A;
                    flag_dp_start = 1;
                    flag_dp_perm = 1;
                    reset_round_cnt = 1;
                end
                else begin
                    next_state = `PERM_B;
                    flag_dp_start = 1;
                    reset_round_cnt = 1;
                end
            end
        end
        `PERM_A : begin
            if (round_cnt == 10) begin
                flag_dpa_transition = 1;
            end
            if(round_cnt == 11) begin
                flag_dpa_done = 1;
                reset_round_cnt = 1;
                flag_dp_start = 1;
                if(flag_dpa_last) begin
                    next_state = `IDLE;
                end
                else if(~flag_dp_perm) begin
                    next_state = `PERM_B;
                end
            end
        end
        `PERM_B : begin
            if (round_cnt == 4) begin
                flag_dpa_transition = 1;
            end
            if(round_cnt == 5) begin
                flag_dpa_done = 1;
                reset_round_cnt = 1;
                flag_dp_start = 1;
                if(flag_dp_perm) begin
                    next_state = `PERM_A;
                end
            end
        end
    endcase
end

always @(posedge clk) begin
    if (flag_dpa_reset) begin
        state <=  `IDLE;
    end
    else begin
        state <= next_state;
    end
end 

always @(posedge clk) begin
    if(reset_round_cnt) begin
        round_cnt <= 0;
    end
    else begin
        round_cnt <= round_cnt + 1;
    end
end

endmodule